#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jan 24 12:04:12 2023

@author: evanqu
"""

# Phlame plot module

#%%
import numpy as np
import os
import pickle
import gzip
import matplotlib.pyplot as plt

import phlame.helper_functions as helper

#%%

class PlotSample():
    '''
    Plot classify data for a single sample.
    '''

    def __init__(self,
                 path_to_frequencies_file,
                 path_to_data_file,
                 path_to_output_plot,
                 max_pi = 0.35,
                 min_prob = 0.5):

        
        self.path_to_frequencies_file = path_to_frequencies_file
        self.path_to_data_file = path_to_data_file
        self.path_to_output_plot = path_to_output_plot
        self.hfont = {'fontname':'Helvetica',
                            'fontsize':12}
        
        self.max_pi = max_pi
        self.min_prob = min_prob

        if not self.path_to_output_plot.endswith('.pdf'):
            self.path_to_output_plot += '.pdf'

        
    def main(self):

        sample_name = os.path.basename(self.path_to_frequencies_file)

        fig = self.plot_sample(sample_name, 
                               self.path_to_frequencies_file,
                               self.path_to_data_file,
                               max_pi = self.max_pi,
                               min_prob = self.min_prob)
        
        fig.tight_layout()

        fig.savefig(self.path_to_output_plot, format='pdf')

  
    def plot_sample(self, sample_name, 
                    path_to_frequencies_file,
                    path_to_data_file,
                    max_pi, min_prob):
        '''
        Plot all classify information for a single sample.
        '''
        
        # Load in sample 
        sample_frequencies = helper.Frequencies(path_to_frequencies_file)
        data = helper.FrequenciesData(path_to_data_file)
        
        # Pick which clades to model
        model_bool = (np.logical_or.reduce((data.counts_MLE != -1),1))
        # model_bool = np.in1d(sample_frequencies.freqs.index , [36, 83])
        
        counts_tmp = []
        total_tmp = []
        clade_names = []
        total_lambda = []
        # ========================================================================
        #  Organize information by clade
        # ========================================================================
        
        for i in range(len(data.clade_counts)):
            
            # Ignore clades with not enough counts to be modeled
            if not model_bool[i]:
                continue
            
            counts_tmp.append( data.clade_counts[i][0] + 
                                data.clade_counts[i][1] )
            
            total_tmp.append( data.clade_counts[i][2] + 
                                data.clade_counts[i][3] )
            
            clade_names.append( sample_frequencies.freqs.index[i] )

            total_lambda.append( data.total_MLE[i][0] )
                

        # ========================================================================
        #  Plot stuff
        # ========================================================================
        
        nplots = np.sum(model_bool)

        if nplots < 2:
            nplots = 2 # Quick fix
        fig, axs = plt.subplots(nplots,4)
        fig.set_size_inches(12, (nplots*2)+2)
        
        # Iterate through clades
        for c, clade in enumerate(clade_names):
            
            # Plot counts histogram
            axs[c,0].set_title(f"{clade}")
            axs[c,0].hist(counts_tmp[c], 
                        bins=np.arange(0,max(counts_tmp[c])+2,1),
                        color='r', alpha=0.5) # csSNP counts
            axs[c,0].hist(total_tmp[c], 
                        bins=np.arange(0,max(total_tmp[c])+2,1),
                        color='k', alpha=0.5) # Total counts
            axs[c,0].set_xlabel('Counts', **self.hfont)
            axs[c,0].set_ylabel('# of SNVs', **self.hfont)
            axs[c,0].tick_params('both', **{'labelsize':12})

            # Bin MCMC chain
            nchain = len(np.array(data.chain)[model_bool][c]['pi'])

            lambda_chain = np.array(data.chain)[model_bool][c]['a']/np.array(data.chain)[model_bool][c]['b']
            pi_chain = np.array(data.chain)[model_bool][c]['pi']

            pi_bins = np.histogram(pi_chain,
                                    np.arange(0,1.01,0.01),
                                    density=True)
            lambda_bins = np.histogram(lambda_chain, 
                                        bins=100,
                                        density=True)
            freq_bins = np.histogram(lambda_chain/data.total_MLE[model_bool][c][0],
                                     np.arange(0,1.01,0.01),
                                     density=True)
            
            
            
            # Calc prob that chain < max_pi
            prob = np.sum(np.array(data.chain)[model_bool][c]['pi']<max_pi)/nchain
            
            # Calc HPD
            hpd = self.get_hpd(np.array(data.chain)[model_bool][c]['pi'], 0.95)

            # Report the output frequency
            
            if prob > min_prob:
                frequency = sample_frequencies.freqs.loc[clade].values[0]
                freq_color='g'
            else:
                frequency = 0
                freq_color='k'
                
            axs[c,0].text(0.95, 0.95, 
            f"Frequency: {frequency:.2f}", 
            ha='right', va='top',
            transform=axs[c,0].transAxes,**{'fontsize':8, 'color':freq_color})

            # pi posterior
            axs[c,1].plot(pi_bins[1][:-1], 
                        pi_bins[0],  
                        color='g', label='Posterior dist.')
            axs[c,1].text(0.99, 0.8, 
                        f"P($\pi$<{max_pi})={prob:.2f}", 
                        ha='right', va='bottom', 
                        transform=axs[c,1].transAxes,**self.hfont)
            axs[c,1].axvline(max_pi, color='k', ls='--', label='max_pi')
            axs[c,1].set_xlabel('$\pi$',**self.hfont); axs[c,1].set_ylabel('Density',**self.hfont)
            axs[c,1].set_xlim(0,1)
            axs[c,1].plot( hpd, [0.1,0.1], color='k', 
                        alpha=0.5, linewidth=4, label='HPD')
            # axs[c,1].legend()
    
            # lambda posterior
            axs[c,2].plot(lambda_bins[1][:-1],
                            lambda_bins[0], 
                            color='b', label='MCMC')
            axs[c,2].set_xlabel('$\lambda$',**self.hfont); axs[c,2].set_ylabel('Density',**self.hfont)
            # axs[c,2].legend()

            # alpha posterior
            axs[c,3].plot(freq_bins[1][:-1],
                          freq_bins[0], 
                          color='b', label='MCMC')
            axs[c,3].set_xlabel('Estimated frequency',**self.hfont); axs[c,3].set_ylabel('Density',**self.hfont)
            
            # Also plot the prior 
                    
        return fig
    
    @staticmethod
    def get_hpd(chain, interval_size=0.95):
        """
        Returns highest probability density region for a given interval
        """
        # Get sorted list
        d = np.sort(np.copy(chain))

        # Number of total samples taken
        n = len(chain)
        
        # Get interval size that should be included in HPD
        interval = np.floor(interval_size * n).astype(int)
        
        # Get width (in units of param) of all intervals 
        int_width = d[interval:] - d[:n-interval]
        
        # Pick out minimal interval
        min_int = np.argmin(int_width)
        
        # Return interval
        return np.array([d[min_int], d[min_int+interval]])


def csSNP_vs_branchlen(tree, csSNPs, clade_names, label):

    names_to_exclude = np.array(['C.1','C.2'])
    exclude_bool = ~np.in1d(clade_names, names_to_exclude)
   
    num_csSNPs = np.zeros(len(clade_names[exclude_bool]))
    branch_lens = np.zeros(len(clade_names[exclude_bool]))

    for i, name in enumerate(clade_names[exclude_bool]):
        
        num_csSNPs[i] = np.count_nonzero(csSNPs[:,exclude_bool][:,i])
        branch_lens[i] = tree.search_nodes(name=name)[0].dist
    
    # branch_lens_sort = np.sort(branch_lens)
    # 
    
    plt.scatter(branch_lens, num_csSNPs, label=label, alpha=.5)
    plt.plot([0,np.max(branch_lens)],[0,np.max(branch_lens)/10],color='k')
    # plt.yscale('log'); plt.xscale('log')
    plt.xlabel('Branch length (# SNPs)')
    plt.ylabel('# csSNPs')
    plt.xlim(0,np.max(branch_lens)+1000)
    plt.ylim(0,np.max(num_csSNPs)+100)

    # fig.show()
    
    return

def nisos_vs_dist_from_tips(tree, 
                            clades, clade_names, 
                            label):
    '''
    Iteratively include clades from tips and plot # of isolates included as a
    fxn of dist from tips.
    '''   
    
    dist_from_tips = np.zeros(len(clade_names))
    
    for i, name in enumerate(clade_names):

        node = tree.search_nodes(name=name)[0]
        
        # Calc average distance from tips
        dist=[]
        for node_leaf in node.iter_leaves():
            dist.append(node.get_distance(node_leaf))
        
        dist_from_tips[i] = np.mean(dist)
        
        
    dist_from_tips_sort = np.sort(dist_from_tips)
    clade_names_sort = clade_names[np.argsort(dist_from_tips)]
    
    isos = []
    nisos = []
    
    for clade in clade_names_sort:
        
        isos = np.unique(np.append(isos,clades[clade]))
        nisos.append( len(isos) )
        
    plt.plot(dist_from_tips_sort, 
              np.array(nisos)/len(tree), label=label)
    # plt.plot(np.arange(0, len(clade_names))/len(clade_names), 
    #           np.array(nisos)/len(tree), label=label)

    plt.ylabel('Percent isolates included')
    plt.xlabel('Average distance from tips (# SNPs)')
    # plt.xlabel('Nth percentile distance from tips')
        
    return


def callable_dist_from_tips(tree, 
                            clade_names,
                            cand_clade_names,
                            lbl):
    '''
    Plot hist of all possible callable clades / clades > 10 csSNPs as a fxn of
    dist from tips
    '''
    
    dist_from_tips = np.zeros(len(cand_clade_names))
    
    for i, name in enumerate(cand_clade_names):

        node = tree.search_nodes(name=name)[0]
        
        # Calc average distance from tips
        dist=[]
        for node_leaf in node.iter_leaves():
            dist.append(node.get_distance(node_leaf))
        
        dist_from_tips[i] = np.mean(dist)
    
    cssnps_bool = np.in1d(cand_clade_names,clade_names)
    
    plt.hist(dist_from_tips, 
             np.logspace(np.log10(1),np.log10(np.max(dist_from_tips)), 40),
             alpha=0.5, color='k', label='candidate clades')
    plt.hist(dist_from_tips[cssnps_bool], 
             np.logspace(np.log10(1),np.log10(np.max(dist_from_tips)), 40),
             alpha=0.5, color='r', label='clades with csSNPs')
    plt.xscale('log')
    plt.xlabel('Average distance from tips (# SNPs)')
    plt.ylabel('Number of clades')
    plt.title(f"{lbl}:  {len(clade_names)/len(cand_clade_names):.2f} candidate clades had csSNPs")
    
    return

def plot_classifier(path_to_classifier):
    
    with gzip.open(path_to_classifier) as f:
        
        cssnp_dct = pickle.load(f)
        
    clade_names = cssnp_dct['clade_names']
    cssnps = cssnp_dct['cssnps']
    
    cssnp_ct = np.zeros(len(clade_names))
    for c in range(len(clade_names)):
        cssnp_ct[c] = (np.count_nonzero(cssnps[:,c]))
        
    # Order decreasing
    cssnp_ct_sort = np.sort(cssnp_ct)
    clade_names_sort = clade_names[np.argsort(cssnp_ct)]
    
    # =========================================================================
    #  Plot
    # =========================================================================
    fig = plt.figure()
    plt.barh(np.arange(0,len(clade_names),1),
             width=cssnp_ct_sort,
             height=0.8, alpha=0.8,
             color='b')
    plt.xlabel('Number of csSNPs'); plt.ylabel('Clade')
    plt.xscale('log')
    plt.yticks(np.arange(0,len(clade_names),1), clade_names_sort)
    plt.xlim(1,max(cssnp_ct_sort)+100)
    # Get number of csssnps to show above each
    for i, v in enumerate(cssnp_ct_sort):
        plt.text(v + 3, i + .25, str(v),
                color = 'k')

    fig.show()
    plt.tight_layout()
    
    return fig

# #%% Re-model a specific clade

# nchain = 10000
# nburn = 500
# seed = 1
# max_pi = 0.3
# min_prob = 0.5
# clade_name = 'F'

# def remodel(sample_name, clade_name,
#             path_to_frequencies_file, path_to_data_file,
#             max_pi, min_prob,
#             nchain, nburn, seed):
    
#     sample_frequencies = Frequencies(path_to_frequencies_file)
#     data = FrequenciesData(path_to_data_file)

#     cladeidx = np.where(sample_frequencies.freqs.index == clade_name)[0][0]

#     counts = data.clade_counts[cladeidx][0] + data.clade_counts[cladeidx][1]
#     total_counts = data.clade_counts[cladeidx][2] + data.clade_counts[cladeidx][3]

#     # Fit the model
#     cts_fit = classify.countsCSS_NEW(counts,
#                                     total_counts,
#                                     seed=seed,
#                                     force_alpha=False)
    
#     prob, hpd = cts_fit.fit(max_pi = max_pi,
#                        nchain = nchain,
#                        nburn = nburn)


#     #################################### Plot ####################################
    
#     fig, axs = plt.subplots(1,4)
    
#     # Plot counts histogram
#     axs[0].hist(counts, 
#                 bins=np.arange(0,max(counts)+2,1),
#                 color='r', alpha=0.5) # csSNP counts
#     axs[0].hist(total_counts, 
#                 bins=np.arange(0,max(total_counts)+2,1),
#                 color='k', alpha=0.5) # Total counts
#     axs[0].set_xlabel('Counts', **self.hfont)
#     axs[0].set_ylabel('# of SNPs', **self.hfont)
#     axs[0].tick_params('both', **{'labelsize':12})


#     # Bin MCMC chain
#     nchain_a = nchain - nburn
#     pi_bins = np.histogram(np.array(cts_fit.chain['pi']),
#                             np.arange(0,1.01,0.01),
#                             density=True)
#     lambda_bins = np.histogram(np.array(cts_fit.chain['a'])/np.array(cts_fit.chain['b']), 
#                                 bins=100,
#                                 density=True)
#     alpha_bins = np.histogram(np.array(cts_fit.chain['a']),
#                                 bins=100,
#                                 density=True)

#     # pi posterior
#     axs[1].plot(pi_bins[1][:-1], 
#                     pi_bins[0],  
#                     color='g', label='Posterior dist.')
#     axs[1].text(0.99, 0.8, 
#                 f"P(pi<{max_pi})={prob:.2f}", 
#                 ha='right', va='bottom', 
#                 transform=axs[1].transAxes,**self.hfont)
#     axs[1].axvline(max_pi, color='g', ls='--', label='max_pi')
#     # axs[c,1].axvline(data.counts_MAP[model_bool][c][2], color='r', label='MAP')
#     axs[1].set_xlabel('Pi',**self.hfont); axs[1].set_ylabel('Probability',**self.hfont)
#     axs[1].set_xlim(0,1)
#     axs[1].plot( hpd, [0.1,0.1], color='k', 
#                 alpha=0.5, linewidth=4, label='HPD')

#     # lambda posterior
#     axs[2].plot(lambda_bins[1][:-1],
#                     lambda_bins[0], 
#                     color='b', label='MCMC')
#     # axs[c,2].axvline(data.counts_MLE[model_bool][c][0], color='k', label='MLE')
#     # axs[c,2].axvline(data.counts_MAP[model_bool][c][0], color='r', label='Posterior mean')
#     axs[2].set_xlabel('Lambda',**self.hfont); axs[2].set_ylabel('Probability',**self.hfont)
#     axs[2].legend()

#     # alpha posterior
#     axs[3].plot(alpha_bins[1][:-1],
#                     alpha_bins[0], 
#                     color='b', label='MCMC')
#     # axs[c,3].axvline(data.counts_MLE[model_bool][c][1], color='k', label='MLE')
#     # axs[c,3].axvline(data.counts_MAP[model_bool][c][1], color='r', label='Posterior mean')
#     axs[3].set_xlabel('Alpha',**self.hfont); axs[3].set_ylabel('Probability',**self.hfont)
#     axs[3].legend()

#     ##############################################################################


#     ################################### Plot 2 ###################################
    
#     fig, axs = plt.subplots(3)
#     fig.set_size_inches(12, 8)
#     # Bin MCMC chain
#     pi_chain = np.array(cts_fit.chain['pi'])
#     lambda_chain = np.array(cts_fit.chain['a'])/np.array(cts_fit.chain['b'])
#     alpha_chain = np.array(cts_fit.chain['a'])
#     nchain = len(pi_chain)

#     # Pi chain
#     axs[0].plot(np.arange(0,nchain), 
#                 pi_chain,  
#                 color='g', label='Posterior dist.')
#     axs[0].set_xlabel('Iteration',**self.hfont); axs[0].set_ylabel('Pi',**self.hfont)
#     axs[0].set_ylim(0,1)
#     axs[0].set_title(f"Pi chain")

#     # lambda posterior
#     axs[1].plot(np.arange(0,nchain),
#                 lambda_chain, 
#                 color='b')
#     axs[1].set_xlabel('Iteration',**self.hfont); axs[1].set_ylabel('Lambda',**self.hfont)
#     axs[1].set_title(f"Lambda chain")


#     # alpha posterior
#     axs[2].plot(np.arange(0,nchain),
#                 alpha_chain, 
#                 color='r')
#     axs[2].set_xlabel('Iteration',**self.hfont); axs[2].set_ylabel('Alpha',**self.hfont)
#     axs[2].set_title(f"Alpha chain")

#     fig.tight_layout()
#     ##############################################################################

#     return fig

# # fig = remodel(sample, '52',
# #             path_to_frequencies_file, path_to_data_file,
# #             max_pi = 0.30,
# #             min_prob = 0.5,
# #             nchain = 10000,
# #             nburn = 500,
# #             seed = 12345)






#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Created on Mon Nov  7 13:39:50 2022

@author: evanqu
"""

#%%
import os
import numpy as np
import pandas as pd
import pickle
import gzip
import warnings
import tempfile
import subprocess
import shlex 
import contextlib
import io

from scipy import stats
from scipy.optimize import minimize 
from statsmodels.base.model import GenericLikelihoodModel
from scipy.special import digamma, loggamma

import phlame.helper_functions as helper

#%%

# Limit for exponentials to avoid overflow
EXP_UPPER_LIMIT = np.log(np.finfo(np.float64).max) - 1.0
        
class Classify:
    '''
    Main controller of the classify step.
    
    Args:
    path_to_cts_file (str): Path to PhLAMe counts file.\n
    
    path_to_classifier (str): Path to PhLAMe classifier file.
    level_input (str): Path to file defining the phylogenetic level.
    to type strains at. This file can either list clade names as 
    they appear in the PhLAMe classifier OR custom group genomes 
    into clades in a 2 column .tsv {genomeID, cladeID}.\n
    
    path_to_frequencies (str): Path to output frequencies file.\n
    
    path_to_data (str, optional): If True, outputs a data file 
    with counts and modeling information at the defined level.
    Defaults to False.\n
    
    max_perc_diff (TYPE, optional): Maximum . Defaults to 0.3.\n
    
    max_snp_diff (TYPE, optional): If True, . Defaults to False.\n
    
    min_snps (int, optional): Minimum number of csSNPs with >0 counts
    to call a lineage as present. Defaults to 10.\n
    
    '''
    
    def __init__(self,
                 path_to_bam,
                 path_to_classifier,
                 ref_file,
                 path_to_frequencies,
                 path_to_cts_file=None,
                 path_to_pileup=None,
                 level_input=False,
                 path_to_data=False,
                 mode='mle',
                 min_snps=10, max_pi=0.3, min_prob=0.5, min_hpd=0.1,
                 nchain=10000, perc_burn=0.1, seed=False, verbose=True):

        self.__path_to_bam_file = path_to_bam
        self.__path_to_pileup = path_to_pileup

        self.__path_to_counts_file = path_to_cts_file
        self.__ref_file = ref_file
        self.__classifier_file = path_to_classifier
        self.__levels_input = level_input
        self.__output_freqs_file = path_to_frequencies
        self.__output_data_file = path_to_data

        self.max_pi = max_pi
        self.min_snps = min_snps
        self.min_prob = min_prob
        self.min_hpd = min_hpd

        self.nchain = nchain
        self.perc_burn = perc_burn
        self.seed = seed

        self.mode = mode

    def main(self):
        
        # =====================================================================
        #  Load Data
        # =====================================================================
        print("Reading in file(s)...")

        self.load_classifier()

        self.load_data()
    
        # =====================================================================
        #  Sort through counts mat to grab just the relevant positions
        # =====================================================================
        print("Sorting counts information...")
        
        self.index_counts()
        
        self.get_allele_counts()

        self.get_counts_alpha()

        # =====================================================================
        #  Calculate & Save frequencies
        # =====================================================================
        print("Modeling counts...")
        
        self.calc_frequencies()
            
        self.save_frequencies()


    def get_positions(self,
                      output_chrpos_file):
        ''' From a classifier object, produce a samtools compatible list of positions
            that are informative to calling clades

        Args:
            path_to_classifier (str): Single path to input classifier file, or path 
                                    to directory with multiple classifier files.\n
            output_allpos_file (str): Path to file giving positions as they appear in
                                    the classifier object (NOT samtools compatible).\n 
            output_chrpos_file (str): Path to samtools compatible list of positions.
            refgenome_folder (str): DESCRIPTION.

        Raises:
            Exception: DESCRIPTION.

        Returns:
            None.

        '''
        cat_pos = np.array([], dtype=np.int32)
        chr_starts, _, scaf_names = helper.genomestats(self.__ref_file)
        
        path_to_cfrs_ls = []
        # Parse whether file or directory of files
        if os.path.isdir(self.__classifier_file):
            
            for filename in os.listdir(self.__classifier_file):
                
                if filename.endswith('.classifier'):
                    path_to_cfrs_ls.append(self.__classifier_file+'/'+filename)
        else:
            path_to_cfrs_ls.append(self.__classifier_file)
            
        
        # Get data from each classifier
        for cfr in path_to_cfrs_ls:
            
            with gzip.open(cfr,'rb') as f:
                csSNPs = pickle.load(f)
                                
                pos = csSNPs['cssnp_pos']
                cat_pos = np.concatenate([pos,cat_pos])

        allpos = np.unique(np.sort(cat_pos))
        
        print(f"{len(allpos)} informative positions found across {len(path_to_cfrs_ls)} classifier(s)")
        
        chr_pos = helper.p2chrpos(allpos, chr_starts)
        chr_names = np.array([scaf_names[i-1] for i in chr_pos[:,0]])
        chr_pos_final = np.vstack((chr_names,chr_pos[:,1])).T
        
        #save as samtools compatible txt file
        # np.savetxt(output_allpos_file, allpos, fmt='%i')
        np.savetxt(output_chrpos_file, chr_pos_final, delimiter='\t', fmt='%s')
        
        return


    def load_classifier(self):
        '''
        Load in classifier and level information
        '''
        
        self.classifier = helper.PhlameClassifier.read_file(self.__classifier_file)
        
        if self.__levels_input:
            self.mylevel = PhyloLevel(self.__levels_input, 
                                      self.classifier.clades, 
                                      self.classifier.clade_names)
            
            # Grab just information for the specific level
            self.level_cfr = self.classifier.grab_level(self.mylevel)
        
        else:
            self.level_cfr = self.classifier

    def load_data(self):

        with tempfile.TemporaryDirectory() as temp_dir:
        
            chrpositions_file = shlex.quote(os.path.join(temp_dir, 'chrpositions.txt'))
            pileup_file = shlex.quote(os.path.join(temp_dir, 'temp.pileup'))
            
            self.get_positions(chrpositions_file)
            
            print("Running samtools mpileup at informative positions...")
            print(f"samtools mpileup -q30 -x -s -O -d3000 "+ \
                           f"-l {chrpositions_file} "+ \
                           f"{shlex.quote(self.__ref_file)} "+ \
                           f"{shlex.quote(self.__path_to_bam_file)} > {pileup_file}")
            
            subprocess.run(f"samtools mpileup -q30 -x -s -O -d3000 "+ \
                           f"-l {chrpositions_file} "+ \
                           f"-f {shlex.quote(self.__ref_file)} "+ \
                           f"{shlex.quote(self.__path_to_bam_file)} > {pileup_file}", shell=True)
            # pileup_file.flush()  # Ensure data is written

            self.countsmat = helper.CountsMat(pileup_file,
                                            self.__ref_file,
                                            self.__classifier_file)
            self.countsmat.main()

    def index_counts(self):
        '''
        Make structures to index through counts array.
        '''
        
        # Informative positions for this level
        # Note this will be > true pos because of pos with multiple alleles
        informative_bool = np.nonzero(self.level_cfr.csSNPs)[0]
        self.informative_pos = self.level_cfr.csSNP_pos[informative_bool]
        
        # Corresponding index on counts mat
        self.informative_pos_idx = np.arange(0,len(self.countsmat.pos))\
            [np.in1d(self.countsmat.pos,self.informative_pos)]\
                [informative_bool]

        # informative_pos_idx = np.arange(0,len(pos))[np.sum(csSNPs,1)>0]
        counts_CSS_bool = np.in1d(self.countsmat.pos,self.informative_pos)

        # Check that all positions are accounted for
        if np.sum(counts_CSS_bool) != len(np.unique(self.informative_pos)):
            raise Exception('Counts and classifier positions do not match. ',
                            'Check that reference genomes are same.')
    
    def get_allele_counts(self):
        
        np.seterr(divide='ignore', invalid='ignore') #ignore divide by 0 warnings
        
        counts = self.countsmat.counts
        counts_idx = self.informative_pos_idx
        alleles = self.level_cfr.alleles
        
        #Initialize data structures
        cts_f = np.zeros(len(counts_idx))
        cts_r = np.zeros(len(counts_idx))
        tot_f = np.zeros(len(counts_idx))
        tot_r = np.zeros(len(counts_idx))
        
        for i, p in enumerate(self.informative_pos):
            
            # -1 +3 is to convert 01234 NATCG to index 
            cts_f[i] = counts[counts_idx[i],
                              int(alleles[i]-1)]
            
            cts_r[i] = counts[counts_idx[i],
                              int(alleles[i]+3)]
            
            tot_f[i] = np.sum(counts[counts_idx[i],:4])
            tot_r[i] = np.sum(counts[counts_idx[i],4:])
            
        self.allele_counts = tuple([cts_f,cts_r,tot_f,tot_r])

    def get_counts_alpha(self):
        '''
        Get overdispersion metric (alpha) across all informative positions.
        '''
        
        counts = self.countsmat.counts
        counts_sum = np.sum(counts, axis=1)

        self.measured_alpha = np.mean(counts_sum)**2/max(1e-6,np.var(counts_sum)-np.mean(counts_sum))
    
    def calc_frequencies(self):

        clade_names = self.level_cfr.clade_names
        nclades = len(clade_names)
        
        frequencies = np.zeros(nclades)
    
        save_cts = []
        save_cts_pos = []
        save_chain = []
        save_hpd = []
        save_cts_map = []

        save_prob=np.full(nclades,-1,dtype=np.float64)
        save_pi=np.full(nclades,-1,dtype=np.float64)
        save_counts_MLE=np.full((nclades,2),-1,dtype=np.float64)
        save_total_MLE=np.full((nclades,2),-1,dtype=np.float64)

        save_flag_highalpha = np.full(nclades,False,dtype=bool)
        save_flag_map_mle_diff = np.full(nclades,False,dtype=bool)

        # Reshape data to group by clade
        for c in range(len(clade_names)):
            
            byclade_cts = self.reshape_byclade(self.allele_counts,
                                               self.level_cfr.allele_cidx, c)
            
            byclade_cts_pos = self.informative_pos[np.where(self.level_cfr.allele_cidx == c)]
    
            cts2model = byclade_cts[0] + byclade_cts[1]
            total2model = byclade_cts[2] + byclade_cts[3]

            # To model, a clade must have a nonzero_zero ratio > 0.03 
            # and at least 10 SNPs with nonzero counts
            if (np.count_nonzero(cts2model) >= self.min_snps):

                with np.errstate(divide='ignore', invalid='ignore'):
                    ratio_nonzero_zero = np.true_divide(np.count_nonzero(cts2model>0),
                                                        np.count_nonzero(cts2model==0))
                    if ratio_nonzero_zero == np.inf:
                        ratio_nonzero_zero = 1.0
                    
                                
                print(f"Fit results for clade: {clade_names[c]}")
                
                fit = countsCSS_NEW(cts2model,
                                    total2model,
                                    seed=self.seed,
                                    mode=self.mode)
                                                    
                frequency, prob = fit.fit(max_pi = self.max_pi,
                                          nchain = self.nchain,
                                          nburn = int(self.perc_burn*self.nchain))

                # print(f"MLE fit: cts lambda={fit.counts_MLE[0]:.2f} cts pi={fit.counts_MLE[1]:.2f}")
                # print(f"total lambda={fit.total_MLE[0]:.2f} total pi={fit.total_MLE[1]:.2f}")
                # print(f"HPD interval: {fit.hpd[0]:.2f}-{fit.hpd[1]:.2f}")
                
                # Only count clades if pass probability and HPD thresholds
                if (prob > self.min_prob) & \
                    (fit.hpd[0] <= self.min_hpd) & \
                        (ratio_nonzero_zero > 0.03):

                    frequencies[c] = fit.frequency
                    
                if fit.measured_alpha > 100000:
                    save_flag_highalpha[c] = True

                # Save fit information
                save_hpd.append(fit.hpd)
                save_cts_map.append(fit.counts_MAP)
                save_counts_MLE[c] = fit.counts_MLE
                save_total_MLE[c] = fit.total_MLE
                save_prob[c] = prob
                save_pi[c] = np.round(fit.pi,4)
                # print(fit.counts_MAP.keys())
                
                # save_chain.append(fit.chain)

            # Otherwise is zero
            else:
                print(f"Clade: {clade_names[c]} ",
                      "does not have not enough SNPs to model")
                # save_chain.append({})
                save_hpd.append(np.array((-1,-1)))
                save_cts_map.append({})

            # Save counts data
            save_cts.append( byclade_cts )
            save_cts_pos.append( byclade_cts_pos )
            
        frequencies_df = pd.DataFrame({'Relative abundance':frequencies,
                                       'DVb':save_pi,
                                       'Probability score':save_prob},
                                      index=self.level_cfr.clade_names)

        self.frequencies = frequencies_df
        self.data = {'clade_counts':save_cts,
                     'clade_counts_pos':save_cts_pos}
        self.fit_info = {'counts_MLE': save_counts_MLE,
                         'total_MLE':save_total_MLE,
                         'counts_MAP':save_cts_map,
                         'chain':[],
                         'prob':save_prob,
                         'hpd':save_hpd}
    
    def save_frequencies(self):
        '''
        Write frequencies and data to files.
        '''        
        
        # Save frequencies
        self.frequencies.to_csv(self.__output_freqs_file, sep=',')
        
        # Save data and fit info
        if self.__output_data_file:
            
            with gzip.open(self.__output_data_file,'wb') as f:
                
                pickle.dump([self.data, self.fit_info],f)

    @staticmethod
    def reshape_byclade(allele_counts, clade_idxs, i):
        '''
        Grab only allele counts belonging to a certain clade (i)
        '''
        
        clade_cts_f = allele_counts[0][np.where(clade_idxs==i)]
        clade_cts_r = allele_counts[1][np.where(clade_idxs==i)]
        clade_tot_f = allele_counts[2][np.where(clade_idxs==i)]
        clade_tot_r = allele_counts[3][np.where(clade_idxs==i)]

        
        return tuple([clade_cts_f,
                      clade_cts_r,
                      clade_tot_f,
                      clade_tot_r])

class countsCSS_NEW:
    '''
    Hold and model counts data covering a single set of cluster-specific SNPs.
    '''
    
    def __init__(self, 
                 counts, total_counts,
                 force_alpha=False,
                 prior=True,
                 prior_strength=20,
                 seed=False,
                 mode='bayesian'):
        
        self.counts = counts
        self.total_counts = total_counts
                
        # Maximum Likelihood fit
        with contextlib.redirect_stdout(io.StringIO()):
            self.counts_MLE = self.zip_fit_mle(self.counts)
            self.total_MLE = self.zip_fit_mle(self.total_counts)
        
        self.force_alpha = force_alpha

        self.prior_strength = prior_strength
        
        self.seed = seed

        self.mode = mode

        # Set hyperparameters
        if force_alpha:
            self.measured_alpha = force_alpha
            
        else:
            self.measured_alpha = np.mean(self.total_counts)**2/max(1e-6,np.var(self.total_counts)-np.mean(self.total_counts))

        m = self.prior_strength # Prior strength; higher values = stronger prior
        logp = m * digamma(self.measured_alpha)
        v = 0; s = 0
        self.params = [m, logp, v, s]

        if not prior:
            self.params = [0, np.log(1), 0, 0]

    def fit(self, max_pi,
            nchain=10000, 
            nburn=500,
            interval_size=0.95,
            subsample=True
            ):
        '''
        Fit counts data to model.
        '''
        
        if self.seed:
            np.random.seed(self.seed)
        
        # =====================================================================
        #  Maximum Likelihood
        # =====================================================================
        if self.mode == 'mle':
        
            lambda_MLE, pi_MLE = self.ZINB_MLE()

            self.counts_MLE = (pi_MLE,lambda_MLE)
            
            self.hpd = np.array([-1,-1])

            self.counts_MAP = {}

            self.prob = int(pi_MLE < max_pi)

            self.frequency = (lambda_MLE/self.total_MLE[0])

            self.pi = pi_MLE


        # =====================================================================
        #  Gibbs sampling
        # =====================================================================
        elif self.mode == 'bayesian':

            lambda_MLE, pi_MLE = self.ZINB_MLE()

            self.counts_MLE = (pi_MLE,lambda_MLE)

            if subsample & (len(self.counts) > 1000):
                self.counts, self.total_counts = self.subsample_positions(1000)
                
            try:
                param_chains, lv_chains = self.ZINB_gibbs_sampler(nchain,
                                                                nburn)
                
            except RuntimeError:
                
                print('Loglikelihood became too low and triggered an overflow error.\
                    Randomly subsampling positions and trying again...')
                
                # subsample_idx = np.random.choice(np.arange(0,len(self.counts)),
                #                                 int(len(self.counts)/2))
                
                # self.counts = self.counts[subsample_idx]
                # self.total_counts = self.total_counts[subsample_idx]

                self.counts, self.total_counts = self.subsample_positions(int(len(self.counts)/2))
                
                param_chains, lv_chains = self.ZINB_gibbs_sampler(nchain,
                                                                nburn)

            
            self.chain = {'pi':param_chains[:,0],
                        'a':param_chains[:,1],
                        'b':param_chains[:,2]}

            self.counts_MAP = {'pi':self.calc_MAP(self.chain['pi']),
                            'a':self.calc_MAP(self.chain['a']),
                            'b':self.calc_MAP(self.chain['b'])}
            
            self.hpd = self.get_hpd(self.chain['pi'], interval_size)
            self.prob = np.sum(self.chain['pi'] < max_pi)/len(self.chain['pi'])

            self.frequency = ((self.counts_MAP['a']/self.counts_MAP['b'])/self.total_MLE[0])

            self.pi = self.calc_MAP(self.chain['pi'])

        else:
            raise ValueError('Mode must be either "mle" or "bayesian"')
            
        return self.frequency, self.prob

    def ZINB_MLE(self):
        
        init = (np.log(np.mean(self.counts)), 0)
        result = minimize(self.zinb_nloglike, init, 
                          args=(self.counts,self.measured_alpha,))

        # Output the optimized parameters
        lambda_MLE, pi_MLE = result.x
        # lambda_ests.append(np.exp(lambda_est))
        # pi_ests.append(1 / (1 + np.exp(-pi_est)))
        # Convert back to bounded values
        return np.exp(lambda_MLE), 1 / (1 + np.exp(-pi_MLE))

    def ZINB_gibbs_sampler(self, nchain, nburn):
        '''
        Gibbs sampling from the following hierarchical model: 
        '''
        nparams = 3 
        npts = len(self.counts)    
        m, logp, v, s = self.params
        
        zero_idx = np.nonzero(self.counts == 0)[0]
        nonzero_idx = np.nonzero(self.counts)[0]    
        
        # Initialize chain
        param_inits, latent_var_inits = self.initialize_chain()
        
        param_samples = np.zeros([nchain, nparams])
        param_samples[0] = param_inits
        latent_var_samples = np.zeros([nchain, 2, npts])
        latent_var_samples[0] = latent_var_inits
    
        for i in range(nchain-1):

            # sample from 𝜆i|...        
            lambda_i = self.sample_lambda_i_conditional(param_samples[i,1], 
                                                       param_samples[i,2], 
                                                       self.counts, 
                                                       latent_var_samples[i,1])
            
            # sample from ri|...
            ri = self.sample_ri_conditional(npts, param_samples[i,0], lambda_i, 
                                           zero_idx, nonzero_idx)
            
            # sample from pi|...
            pi = self.sample_p_conditional(ri, npts)
            
            # sample from b|...
            b = max(1e-6,self.sample_b_conditional(param_samples[i,1], v, s, lambda_i, npts))
            
            # sample from a|...
            a_chain = self.sample_a_conditional_gp([lambda_i,b,v,m,logp], 
                                                    param_samples[i,1])
            a = a_chain[-1]
            
            param_samples[i+1] = np.array([pi, a, b])
            latent_var_samples[i+1] = np.array([lambda_i,ri])
        
        # Convert pi into percent zero-inflated
        param_samples[:,0] = (1 - param_samples[:,0])
        
        return param_samples, latent_var_samples
    
    def initialize_chain(self):
        '''
        Initialize parameters and latent variables for Gibbs sampling.
        '''
        
        counts_mean = self.counts.mean()
        
        pi_init = min(1 - ((self.counts == 0).mean() - stats.poisson.pmf(0, counts_mean)),0.99)
        a_init = self.measured_alpha
        b_init = a_init/counts_mean
        
        param_inits = np.array([pi_init, a_init, b_init])
        
        
        ri_init = np.int64(self.counts > 0)
        predicted_zeros = (1/(1+self.measured_alpha*counts_mean))**(1/self.measured_alpha)
        ri_init[ri_init == 0] = stats.bernoulli.rvs(predicted_zeros, size=np.sum(ri_init==0))
        lambda_init = [self.counts.mean()]*len(self.counts)

        latent_var_inits = np.array([lambda_init,ri_init])
        
        return param_inits, latent_var_inits

    def subsample_positions(self, n):
        '''
        Subsample n positions from the counts data.
        '''
        
        idx = np.random.choice(np.arange(0,len(self.counts)), n)
        
        return self.counts[idx], self.total_counts[idx]
        
    def zip_fit_mle(self, cts):
        
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            model = ZeroInflatedPoisson(cts)
            fit = model.fit()
            pi, lambda_ = fit.params
        
        # return lambda_, pi
            # CI_pi, CI_lambda_ = fit.conf_int()
            # # range_CI_pi = CI_pi[1] - CI_pi[0]
            # range_CI_lambda_ = CI_lambda_[1] - CI_lambda_[0]
        
        return np.array([lambda_, pi])
        
    @staticmethod
    def zinb_nloglike(params, data, alpha_prior):
        
        # Unpack the real-valued input params
        x_lambda, x_pi = params
        
        # Apply transformations to get the actual parameters
        lambd = np.exp(x_lambda)  # Lambda must be positive
        pi = 1 / (1 + np.exp(-x_pi))  # Pi must be in the interval (0, 1)
        
        # Calculate the probability of success p using lambda and alpha
        p = alpha_prior / (lambd + alpha_prior)
        
        # Create a boolean mask for zero observations
        zero_mask = (data == 0)
        
        # Calculate the Negative Binomial PMF for all y
        nb_pmf = stats.nbinom.pmf(data, alpha_prior, p)
        
        # Compute log likelihood for zeros
        log_likelihood_zeros = np.log(pi + (1 - pi) * nb_pmf)  # Log for the zeros
        
        # Compute log likelihood for non-zeros
        log_likelihood_nonzeros = np.log((1 - pi) * nb_pmf)  # Log for the non-zeros
        
        # Apply the appropriate log likelihood based on whether y = 0 or y > 0
        log_likelihood = np.sum(zero_mask * log_likelihood_zeros + (1 - zero_mask) * log_likelihood_nonzeros)
        
        return -log_likelihood

    @staticmethod
    def calc_MAP(chain, bins=False):

        if not bins:
            bins = 100

        counts, bins = np.histogram(chain, bins=bins)
        max_idx = np.argmax(counts)
        
        return bins[max_idx]
    
    @staticmethod
    def _calc_p_posterior(total_counts, scaleby,
                          b=1,
                          c=1):
        '''
        Empirical posterior distribution over p for Nbinom-distributed counts.
        Calculate with a beta(b,c) prior over p
        '''
        
        #If underdispersed relative to Poisson skip calc because it will break
        if np.var(total_counts) <= np.mean(total_counts):
            
            return (np.sum(total_counts),1)
            
        k = np.mean(total_counts)**2/(np.var(total_counts)-np.mean(total_counts))
        T = len(total_counts)
        alpha_update = (k * T) + b
        beta_update = np.sum(total_counts) + c
        
        # Empirical max value of p
        # p_max = np.mean(total_counts)/np.var(total_counts)
        
        return (alpha_update/scaleby, beta_update/scaleby)

    
    @staticmethod
    def _zip_nloglike(params, counts):
        '''
        Negative log-likelihood function for zero-inflated Poisson
        '''
        
        lambda_ = params[0]
        pi = params[1]
        
        if lambda_ <= 0:
            return np.inf
        if pi < 0 or pi > 1:
            return np.inf
    
        Y=len(counts) - np.count_nonzero(counts)
        n=len(counts)
        
        return -( Y*np.log(pi + (1 - pi)*np.exp(-lambda_)) +
                  (n - Y) * np.log(1 - pi) - 
                  (n - Y) * lambda_ + 
                  n * np.mean(counts) * np.log(lambda_) -
                  np.sum(np.log(np.array([np.math.factorial(int(c)) for c in counts], dtype=float))) )
                  #np.log(np.product(np.array([np.math.factorial(int(c)) for c in counts], dtype='float128'))) )
    
    @staticmethod
    def sample_lambda_i_conditional(a, b, counts, r_i):
        
        return stats.gamma.rvs(a = a+counts,
                               scale = 1/(r_i+b))
    
    @staticmethod
    def sample_p_conditional(ri, npts):
        
        return stats.beta.rvs(np.sum(ri)+1, npts - np.sum(ri) + 1)
    
    @staticmethod
    def sample_ri_conditional(npts, pi, lambda_i, 
                              zero_idx, nonzero_idx):
        # If yi > 0, then point i belongs to the at-risk class,
        # and hence by definition, ri = 1 with probability 1.
        # If yi = 0, then we observe either a structural zero (ri = 0) 
        # or an at-risk zero (ri = 1). 
        
        # Draw wi from a Bernoulli distribution with probability
        ri = np.zeros(npts)
        
        predicted_prob_zero = (1/(1+(np.var(lambda_i)/np.mean(lambda_i))))**(np.mean(lambda_i)**2/np.var(lambda_i))
        
        # zero_bern = ( (pi*np.exp(-np.sum(lambda_i)))/
        #               ((pi*np.exp(-np.sum(lambda_i)))+(1-pi)) )
        zero_bern = ( (pi*predicted_prob_zero)/
                      ((pi*predicted_prob_zero)+(1-pi)) )
    
        # Note as np.exp(-np.sum(lambda_i)) reaches the floating
        # point limit the overall term will go to 0
        if len(zero_idx) > 0:
            ri[zero_idx] = stats.bernoulli.rvs(zero_bern, size=len(zero_idx))
        if len(nonzero_idx) > 0:
            ri[nonzero_idx] = stats.bernoulli.rvs(1, size=len(nonzero_idx))
    
        return ri
    
    @staticmethod
    def sample_a_conditional_gp(aparams,
                                a_old):
        
        def log_a_conditional_dist_gp(a, params):
            ''' 
            Conditional distribution of a (gamma shape parameter) 
            in a gamma-poisson model
            '''
            
            lambda_i, b, v, m, logp = params
            
            # Will cause -inf error
            lambda_i = lambda_i[np.nonzero(lambda_i)]
            
            npts = len(lambda_i)
            
            return (a*(npts+v)*np.log(b) + 
                    (a)*logp + 
                    a*np.sum(np.log(lambda_i)) - 
                    (m + npts)*loggamma(a))

        if log_a_conditional_dist_gp(a_old, aparams) == -np.inf:
            raise RuntimeError('Loglikelihood is too low!')
            
        chain = slice_sampler(x0=a_old, 
                              loglike=log_a_conditional_dist_gp, 
                              params=aparams,
                              niter=50,
                              sigma=a_old/10, 
                              step_out=True)
        
        return chain
    
    @staticmethod
    def sample_b_conditional(a,
                             v, s,
                             data, npts):
        
        return stats.gamma.rvs(a = a*(npts + v),
                                scale = 1/(s + np.sum(data)))    
    
    
    @staticmethod
    def get_hpd(chain, interval_size=0.95):
        """
        Returns highest probability density region for a given interval
        """
        # Get sorted list
        d = np.sort(np.copy(chain))
    
        # Number of total samples taken
        n = len(chain)
        
        # Get interval size that should be included in HPD
        interval = np.floor(interval_size * n).astype(int)
        
        # Get width (in units of param) of all intervals 
        int_width = d[interval:] - d[:n-interval]
        
        # Pick out minimal interval
        min_int = np.argmin(int_width)
        
        # Return interval
        return np.array([d[min_int], d[min_int+interval]])
    
def slice_sampler(x0, 
                  loglike, params,
                  niter, sigma, 
                  step_out=True):
    """
    based on http://homepages.inf.ed.ac.uk/imurray2/teaching/09mlss/
    """

    # set up empty sample holder
    samples = np.zeros(niter)

    # initialize
    xx = float(x0)

    last_llh = loglike(xx, params)

    for i in range(niter):
                    
        # =================================================================
        #  Randomly sample y from (0,L(X))
        # =================================================================
        # aka multiply likelihood by uniform[0,1]
        llh0 = last_llh + np.log(np.random.rand())
        
        # =================================================================
        #  Get straight line segment at y under the curve L(x)
        # =================================================================
        rr = np.random.rand(1)

        # Initialize left and right of segment with total distance sigma
        x_l = float(xx)
        x_l = x_l - rr * sigma
        x_r = float(xx)
        x_r = x_r + (1 - rr) * sigma
        
        # Continue to take steps of sigma out until you get ends outside the curve
            #"stepping out"
        if step_out:
            llh_l = loglike(x_l, params)
            while llh_l > llh0:
                x_l = x_l - sigma
                llh_l = loglike(x_l, params)
            llh_r = loglike(x_r, params)
            while llh_r > llh0:
                x_r = x_r + sigma
                llh_r = loglike(x_r, params)

        # =================================================================
        #  Randomly sample a new x from within the total segment 
        # =================================================================
            #"shrinkage"
        x_cur = float(xx)
        while True:
            xd = np.random.rand() * (x_r - x_l) + x_l
            x_cur = float(xd)
            last_llh = loglike(x_cur, params)
            
            #If the loglikelihood of new x is higher than y, good to go
            if last_llh > llh0:
                xx = float(xd)
                break
            # Otherwise make new x the new left/right end of segment
            elif xd > xx:
                x_r = xd
            elif xd < xx:
                x_l = xd
            else:
                raise RuntimeError('Slice sampler shrank too far.')

        samples[i] = float(xx)

    return samples

class ZeroInflatedPoisson(GenericLikelihoodModel):
    
    def __init__(self, endog, exog=None, **kwds):
        if exog is None:
            exog = np.zeros_like(endog)
            
        super(ZeroInflatedPoisson, self).__init__(endog, exog, **kwds)
    
    def nloglikeobs(self, params):
        pi = params[0]
        lambda_ = params[1]

        return -np.log(self._zip_pmf(self.endog, pi=pi, lambda_=lambda_))
    
    def fit(self, start_params=None, maxiter=10000, maxfun=5000, **kwds):
        if start_params is None:
            lambda_start = self.endog.mean()
            excess_zeros = (self.endog == 0).mean() - stats.poisson.pmf(0, lambda_start)
            pi_start = excess_zeros if excess_zeros>0 else 0
            start_params = np.array([pi_start, lambda_start])
            
        return super(ZeroInflatedPoisson, self).fit(start_params=start_params,
                                                    maxiter=maxiter, maxfun=maxfun, **kwds)
    
    @staticmethod
    def _zip_pmf(x, pi, lambda_):
        '''zero-inflated poisson function, pi is prob. of 0, lambda_ is the fit parameter'''
        if pi < 0 or pi > 1 or lambda_ <= 0:
            return np.zeros_like(x)
        else:
            return (x == 0) * pi + (1 - pi) * stats.poisson.pmf(x, lambda_)

class PhyloLevel:
    '''
    Holds a set phylogenetic level in reference to a classifier.
    '''
    
    def __init__(self, levelin, allclades, allclade_names):
        
        # Import information about the tree
        # Note: ideally this will be replaced by an import of a classifier class
        self.__allclades = allclades
        self.__allclade_names = allclade_names
        
        # Import levels from file or string
        if os.path.isfile(levelin):
            print('Reading in levels file...')
            self.clades, self.clade_names, self.names = self._parse_file(levelin)
        
        else: 
            print('Searching for the following levels in classifier:')
            print(levelin)
            self.clades, self.clade_names, self.names = self._parse_str(levelin)
            
    def _parse_file(self, levelin_file, uncl='-1'):
        '''
        Parse a 2 column delimited levels file 
        ( e.g genome_name1, clade_ID1
              genome_name2, clade_ID2 )
        '''
        
        with open(levelin_file,'r') as file:
            firstline = file.readline()
       
        if len(firstline.strip().split('\t'))==2:
            dlim='\t'
        elif len(firstline.strip().split(','))==2:
            dlim=','


        clade_ids = np.loadtxt(levelin_file, delimiter=dlim, dtype=str)
        
        clades, names = self._reshape(clade_ids[:,0],
                                      clade_ids[:,1],
                                      '-1')
        
        # Check that level is valid within classifier
        clade_names = self._check_file(clades, names)
        
        # Check that levels are not direct ancestors or descendants
        # of each other.
        self._ancdesc(clades, clade_names)
        
        return clades, clade_names, names

    def _parse_str(self, levelin_str):
        '''
        Parse a comma-delimited list of clade names (e.g. C.1,C.2,C.3)
        '''
        clade_names = levelin_str.strip().split(',')
        
        # Check that level is valid within classifier
        clades = self._check_str(clade_names)
        names = clade_names
        
        # Check that levels are not direct ancestors or descendants
        # of each other.
        self._ancdesc(clades, clade_names)
        
        return clades, clade_names, names
    
    def _check_file(self, clades, names):
        '''
        Check that groupings specified by a file are valid clades in classifer.
        '''
        
        clade_names=[]
        allclades_ls = list(self.__allclades.values())
        allclade_names = np.array(list(self.__allclades.keys()))
        
        # Plus return the missing data (clade_names)
        for i, clade in enumerate(clades):
        
            # Check if grouping of genomes is actually a clade in classifier
            match_bool = [ set(clade) == set(genomes) for genomes in allclades_ls ]
            
            if np.sum(match_bool)==0:
                raise Exception(f"Error: Could not find a valid clade corresponding to {names[i]} in classifier.")
            
            # Ugly
            clade_names.append(str(allclade_names[match_bool][0]))
            
        return clade_names
    
    def _check_str(self, clade_names):
        '''
        Check if clade names specified are valid clades in classifier.
        '''
        clades = []
        
        for name in clade_names:

            if name not in self.__allclade_names:
                raise Exception(f"Error: {name} is not a valid clade in the classifier!")

            clades.append(self.__allclades[name])
            
        return clades
    
    @staticmethod
    def _ancdesc(clades, clade_names):
        '''
        Check that no two clades in a level are direct ancestors or descendants 
        of each other.
        '''

        for i, clade in enumerate(clades):
            
            # Get other clades to cp against
            cp = (clades[:i] + clades[i+1 :])
            
            # Are any genomes duplicated in 2 clades
            dup_bool = [len(set(clade).intersection(set(genomes)))>0 for genomes in cp]

            if np.sum(dup_bool) > 0:
                raise Exception(f"Error: {clade_names[i]} is either an ancestor or descendant of {clade_names[dup_bool.index(True)]}.",
                                "Note that within a level the same genome cannot be included in two clades.")

    @staticmethod
    def _reshape(names_long, clade_ids_long, uncl_marker):
        '''
        Reshape a two column 'long' levels array into list of arrays
        '''
        
        clades = []; names = []
        
        for c in np.unique(clade_ids_long):
            names.append(str(c))
            clades.append(names_long[np.in1d(clade_ids_long,c)])
        
        # Remove things designated as unclassifed
        if uncl_marker in names:
            # Get index of unclassified names
            idx = names.index(uncl_marker)
            # And remove
            del names[idx]; del clades[idx]
        
        else:
            print('Nothing was found as unclassified in clade IDs. Ignore if intentional!')
                
        return clades, names



# -*- coding: utf-8 -*-
'''

'''

import numpy as np
import gzip
import os
import pickle
import phlame.helper_functions as helper

#%% combine_positions.py

class CombinePositions():

    def __init__(self,
                 path_to_diversity_files,
                 path_to_ref):
        
        self.diversity_files = helper.parse_file_list(path_to_diversity_files)
        self.path_to_ref = path_to_ref

        self.chr_starts, self.genome_length, self.scaf_names = helper.genomestats(self.path_to_ref)

    def main(self):
        
        variant_position_ls = []
        for file_ in self.diversity_files:

            with gzip.open(file_, 'rb') as f:
                div = pickle.load(f)
                variant_pos = div['variant_pos']
            variant_position_ls.append(variant_pos)

        all_positions = self.combine_positions(variant_position_ls)

        self.all_positions = all_positions

        return all_positions

    def combine_positions(self,
                          positions_files_ls):
        
        #in_outgroup: booleans giving whether or not each sample is in the outgroup
        # print("Processing outgroup booleans...")
        # in_outgroup=[]
        # with open(path_to_outgroup_boolean_file) as file:
        #     for line in file:
        #         in_outgroup.append(line)
        # #Bool of samples to include
        # include = [not i for i in in_outgroup]
        
        # #Get positions on reference genome
        # [self.chr_starts,self.genome_length,self.scaf_names] = ghf.genomestats(REFGENOMEDIRECTORY)
        
        #Find positions with at least 1 fixed mutation relative to reference genome
        print('\n\nFinding positions with at least 1 fixed mutation...\n')
        
        # positions_files_ls=[]
        # with open(path_to_positions_files) as file:
        #     for line in file:
        #         positions_files_ls.append(line)
                
        
        cp = self.generate_positions_snakemake(positions_files_ls)
        print(f"Found {len(cp)} positions where provided vcfs called a fixed variant in at least one in-group sample \n")

        #Todo: Add candidate positions manually
        #Combine different types of positions
        allp = cp
                
        return allp

    def generate_positions_single_sample(self,
                                         path_to_variant_vcf,
                                         maxFQ=-30):
        '''
        Python version of generate_positions_single_sample_snakemake.m

        Args:
            path_to_variant_vcf (str): Path to .variant.vcf.gz file.
            maxFQ (int): Purity threshold for including position.
            outgroup_bool (bool): Whether this sample is outgroup or not.

        Returns:
            None.

        '''    
        # print(f"Currently examining the following vcf file: {path_to_variant_vcf}\n")
        # print(f"FQ threshold: {int(maxFQ)}")
        
        # Initialize boolean vector for positions to include as candidate SNPs that
        # vary from the reference genome
        include = np.zeros((self.genome_length,1))
        
        
        f = gzip.open(path_to_variant_vcf,'rt')
        
        for line in f:
            if not line.startswith("#"):
                lineinfo = line.strip().split('\t')
                
                chromo=lineinfo[0]
                position_on_chr=lineinfo[1] #1-indexed
                
                if len(self.chr_starts) == 1:
                    position=int(lineinfo[1])
                else:
                    if chromo not in self.scaf_names:
                        raise ValueError("Scaffold name in vcf file not found in reference")
                    position=int(self.chr_starts[np.where(chromo==self.scaf_names)]) + int(position_on_chr)
                    #self.chr_starts begins at 0
                    
                alt=lineinfo[4]
                ref=lineinfo[3]
                
                #only consider for simple calls (not indel, not ambiguous)
                if (alt) and ("," not in alt) and (len(alt) == len(ref)) and (len(ref)==1):
                    #find and parse quality score
                    xt = lineinfo[7]
                    xtinfo = xt.split(';')
                    entrywithFQ=[x for x in xtinfo if x.startswith('FQ')][0]
                    fq=entrywithFQ[entrywithFQ.index("=")+1:]
                    
                    if float(fq) < maxFQ: #better than maxFQ
                        include[position-1]=1
                        #-1 converts position (1-indexed) to index
        
        #+1 converts index back to position for p2chrpos
        var_positions=helper.p2chrpos(np.nonzero(include)[0]+1,self.chr_starts)
        
        #save
        # with gzip.open(path_to_output_positions,"wb") as f:
        #     pickle.dump(Var_positions,f)
            
        # print(f"{len(Var_positions)} variable positions found passing quality threshold")
        
        return var_positions


    def chrpos2index(self, chrpos):
        '''Python version of chrpos2index.m

        Args:
            chrpos (arr): px2 array of position and chromsome idx.
            self.chr_starts (arr): Vector of chromosome starts (begins at 0).

        Returns:
            p (arr): Vector of position indexes.

        '''
        if np.size(chrpos,0) < np.size(chrpos,1):
            chrpos=chrpos.T
            print('Reversed orientation of chrpos')
            
        if len(self.chr_starts) == 1:
            p=chrpos[:,1]
        else:
            p=self.chr_starts[chrpos[:,0]-1]+chrpos[:,1]

        return p


    def generate_positions_snakemake(self, positions_list):
        '''Python version of generate_positions_snakemake.m
        
        Args:
            paths_to_input_p_files (list): List of input positions files.
            REFGENOMEDIRECTORY (str): Path to reference genome.

        Returns:
            combined_pos (arr): Vector of variable positions across samples.

        '''
                
        # initialize vector to count occurrances of variants across samples
        timesvariant = np.zeros((self.genome_length,1))
        
        # for i in range(len(positions_files_list)):
        #     #load in positions array for sample
        #     with gzip.open(positions_files_list[i].rstrip('\n'),"rb") as f:
        #     # with open(positions_files_list[i].rstrip('\n'),"rb") as f:
        #         positions=pickle.load(f)
            
        #     if len(positions)>2:
        #         x=self.chrpos2index(positions)
                
        #         timesvariant[x]=timesvariant[x]+1

        for positions in positions_list:

            if len(positions) > 2:
                x = self.chrpos2index(positions)
                timesvariant[x] = timesvariant[x] + 1
        
        
        #Keep positions that vary from the reference in at least one sample but
        #that don't vary from the reference in ALL samples
        combined_pos = np.where((timesvariant > 0) & (timesvariant < len(positions_list)))[0]
        
        return combined_pos

#%% pileup2diversity.py

class Pileup2Diversity:

    def __init__(self, 
                 path_to_pileup,
                 path_to_vcf,
                 path_to_variant_vcf,
                 path_to_ref,
                 path_to_output_diversity=None):
        
        self.path_to_pileup = path_to_pileup
        self.path_to_vcf = path_to_vcf
        self.path_to_variant_vcf = path_to_variant_vcf

        self.path_to_output_diversity = path_to_output_diversity

        self.chr_starts, self.genome_length, self.scaf_names = helper.genomestats(path_to_ref)

    def main(self):

        data, coverage = self.pileup2diversity(self.path_to_pileup,
                                               self.chr_starts,
                                               self.genome_length,
                                               self.scaf_names)
        

        quals_sample = Vcf2Quals.vcf_to_quals(self.path_to_vcf,
                                              self.chr_starts,
                                              self.genome_length,
                                              self.scaf_names)
        
        variant_pos = self.generate_positions_single_sample(self.path_to_variant_vcf)
        
        # print(variant_pos)

        if self.path_to_output_diversity:

            self.write_diversity(data, quals_sample, variant_pos,
                                 self.path_to_output_diversity)

    @staticmethod
    def write_diversity(data,
                        quals_sample,
                        variant_pos, path_to_output_diversity):
        
        diversity = {'data': data,
                     'quals': quals_sample,
                     'variant_pos': variant_pos}
        
        print(diversity['variant_pos'])

        with gzip.open(path_to_output_diversity, 'wb') as f:
            
            pickle.dump(diversity, f)

    @staticmethod
    def pileup2diversity(path_to_pileup,
                         chr_starts, genome_length, scaf_names):
        """Grabs relevant allele info from mpileupfile and stores as a nice array
        This version ONLY records counts and indel_counter

        Args:
            input_pileup (str): Path to input pileup file.
            path_to_ref (str): Path to reference genome file
            
        """
        #parameters
        nts = 'ATCGatcg'
        nts_dict = {'A':0,'T':1,'C':2,'G':3,'a':4,'t':5,'c':6,'g':7}
        num_fields=10
        indelregion=3 #region surrounding each p where indels recorded 
        #get reference genome + position information
        
        #initialize output array
        data = np.zeros((genome_length,num_fields)) #format [[A T C G  a t c g],[...]]
        
        #read in mpileup file
        print(f"Reading input file: {path_to_pileup}")
        mpileup = open(path_to_pileup)
        
        #####
        loading_bar=0
        
        for line in mpileup:
            
            # loading_bar+=1
            # if loading_bar % 50000 == 0:
            #     print('.')
            
            lineinfo = line.strip().split('\t')
            
            #holds info for each position before storing in data
            temp = np.zeros((num_fields))
            
            chromo = lineinfo[0]
            #position (absolute)
            if len(chr_starts) == 1:
                position=int(lineinfo[1])
            else:
                if chromo not in scaf_names:
                    raise ValueError("Scaffold name in pileup file not found in reference")
                position=int(chr_starts[np.where(chromo==scaf_names)]) + int(lineinfo[1])
                #chr_starts begins at 0
            
            #ref allele
            ref=nts_dict[lineinfo[2]] # convert to 0123
            if ref > 4:
                ref = ref - 4
            
            #calls info
            #calls=lineinfo[4]
            calls=np.fromstring(lineinfo[4], dtype=np.int8) #to ASCII
            # spits out a warning
            # calls=np.array([ord(l) for l in lineinfo[4]]) #ASCII
            
            #find starts of reads ('^' in mpileup)
            startsk=np.where(calls==94)[0]
            for k in startsk:
                calls[k:k+2]=-1 #WHAT IS -1
                #remove mapping character, absolutely required because the next chracter could be $
            
            #find ends of reads ('$' in mpileup)
            endsk=np.where(calls==36)[0]
            calls[endsk]=-1
            
            #find indels + calls from reads supporting indels ('+-')
            indelk = np.where((calls==43) | (calls==45))[0]
            for k in indelk:
                if (calls[k+2] >=48) and (calls[k+2] < 58): #2 digit indel (size > 9 and < 100)
                    indelsize=int(chr(calls[k+1]) + chr(calls[k+2])) 
                    #indelsize=str2double(char(calls(k+1:k+2))); MATLAB
                    indeld=2
                else: #1 digit indel (size <= 9)
                    indelsize=int(chr(calls[k+1]))
                    indeld=1
                #record that indel was found in +/- indelregion nearby
                #indexing is slightly different here from matlab version
                if calls[k]==45: #deletion
                    if (position-indelregion-1 >= 0) and (position+indelsize+indelregion-1 < genome_length):
                        #must store directly into data as it affects lines earlier and later
                        data[position-indelregion-1:position+indelsize+indelregion-1,9]+=1
                    elif position-indelregion >= 0: #for indels at end
                        data[position-indelregion-1:,9]+=1
                    else: #for indels at beg
                        data[:position+indelsize+indelregion-1,9]+=1
                else: #insertion
                    #insertion isn't indexed on the chromosome, no need for complex stuff
                    if (position-indelregion-1 >= 0) and (position+indelregion-1 < genome_length):
                        data[position-indelregion-1:position+indelregion-1,8]+=1
                    elif position-indelregion >= 0:
                        data[position-indelregion-1:,8]+=1
                    else:
                        data[:position+indelsize-1,8]+=1

                #remove indel info from counting
                calls[k:(k+1+indeld+indelsize)] = -1 #don't remove base that precedes an indel
            
            #replace reference matches (.,) with their actual calls
            if ref >=0:
                calls[np.where(calls==46)[0]]=ord(nts[ref]) #'.'
                calls[np.where(calls==44)[0]]=ord(nts[ref+4]) #','

            #index reads for finding scores
            simplecalls=calls[np.where(calls>0)[0]]
            #simplecalls is a tform of calls where each calls position
            #corresponds to its position in bq, mq, td
            
            #count how many of each nt and average scores

            for nt in range(8):
                nt_count=np.count_nonzero(simplecalls == ord(nts[nt]))
                if nt_count > 0:
                    temp[nt]=nt_count
            
            #-1 is needed to turn 1-indexed mpileup to 0-indexed arr
            data[position-1,0:8]=temp[0:8]
            
        #######
        mpileup.close()
        
        #calc coverage
        coverage=np.sum(data[:,0:8],1)
        
        return data, coverage
    
    def generate_positions_single_sample(self,
                                         path_to_variant_vcf,
                                         maxFQ=-30):
        '''
        Python version of generate_positions_single_sample_snakemake.m

        Args:
            path_to_variant_vcf (str): Path to .variant.vcf.gz file.
            maxFQ (int): Purity threshold for including position.
            outgroup_bool (bool): Whether this sample is outgroup or not.

        Returns:
            None.

        '''    
        # print(f"Currently examining the following vcf file: {path_to_variant_vcf}\n")
        # print(f"FQ threshold: {int(maxFQ)}")
        
        # Initialize boolean vector for positions to include as candidate SNPs that
        # vary from the reference genome
        include = np.zeros((self.genome_length,1))
        
        
        f = gzip.open(path_to_variant_vcf,'rt')
        
        for line in f:
            if not line.startswith("#"):
                lineinfo = line.strip().split('\t')
                
                chromo=lineinfo[0]
                position_on_chr=lineinfo[1] #1-indexed
                
                if len(self.chr_starts) == 1:
                    position=int(lineinfo[1])
                else:
                    if chromo not in self.scaf_names:
                        raise ValueError("Scaffold name in vcf file not found in reference")
                    position=int(self.chr_starts[np.where(chromo==self.scaf_names)]) + int(position_on_chr)
                    #self.chr_starts begins at 0
                    
                alt=lineinfo[4]
                ref=lineinfo[3]
                
                #only consider for simple calls (not indel, not ambiguous)
                if (alt) and ("," not in alt) and (len(alt) == len(ref)) and (len(ref)==1):
                    #find and parse quality score
                    xt = lineinfo[7]
                    xtinfo = xt.split(';')
                    entrywithFQ=[x for x in xtinfo if x.startswith('FQ')][0]
                    fq=entrywithFQ[entrywithFQ.index("=")+1:]
                    
                    if float(fq) < maxFQ: #better than maxFQ
                        include[position-1]=1
                        #-1 converts position (1-indexed) to index
        
        #+1 converts index back to position for p2chrpos
        var_positions=helper.p2chrpos(np.nonzero(include)[0]+1,self.chr_starts)
        
        #save
        # with gzip.open(path_to_output_positions,"wb") as f:
        #     pickle.dump(Var_positions,f)
            
        # print(f"{len(Var_positions)} variable positions found passing quality threshold")
        
        return var_positions


class Vcf2Quals:

    @staticmethod
    def vcf_to_quals(path_to_vcf_file,
                     chr_starts, genome_length, scaf_names):
        '''
        Python version of vcf_to_quals_snakemake.py
        Given a vcf file with one file per line, grabs FQ score for each positions. Ignores lines corresponding to indels

        Args:
            path_to_vcf_file (str): Path to .vcf file.
            output_path_to_quals (str): Path to output quals file
            REFGENOMEDIRECTORY (str): Path to reference genome directory.

        Returns:
            None.
            
        '''

        #initialize vector to record quals
        quals = np.zeros((genome_length,1), dtype=int)
        
        print(f"Loaded: {path_to_vcf_file}")
        file = gzip.open(path_to_vcf_file,'rt') #load in file
        
        for line in file:
            if not line.startswith("#"):
                lineinfo = line.strip().split('\t')
                
                #Note: not coding the loading bar in the matlab script
                
                chromo=lineinfo[0]
                position_on_chr=lineinfo[1] #1-indexed
                
                if len(chr_starts) == 1:
                    position=int(lineinfo[1])
                else:
                    if chromo not in scaf_names:
                        raise ValueError("Scaffold name in vcf file not found in reference")
                    position=int(chr_starts[np.where(chromo==scaf_names)]) + int(position_on_chr)
                    #chr_starts begins at 0
                    
                alt=lineinfo[4]
                ref=lineinfo[3]
                
                #only consider for simple calls (not indel, not ambiguous)
                if (alt) and ("," not in alt) and (len(alt) == len(ref)) and (len(ref)==1):
                    #find and parse quality score
                    xt = lineinfo[7]
                    xtinfo = xt.split(';')
                    entrywithFQ=[x for x in xtinfo if x.startswith('FQ')][0]
                    fq=float(entrywithFQ[entrywithFQ.index("=")+1:])
                    
                    #If already a position wiht a stronger FQ here, don;t include this
                    #More negative is stronger
                    if fq < quals[position-1]:
                        quals[position-1]=round(fq) 
                            #python int(fq) will by default round down, round matches matlab behavior
                            #-1 important to convert position (1-indexed) to python index
        
        return quals

#%%

class Case():

    def __init__(self,
                 path_to_sample_names,
                 path_to_diversity_files,
                 path_to_ref,
                 path_to_out_cmt):
        
        self.path_to_sample_names_file = path_to_sample_names
        
        self.path_to_diversity_files = path_to_diversity_files
        self.diversity_files = helper.parse_file_list(self.path_to_diversity_files)

        self.path_to_out_cmt = path_to_out_cmt
        self.path_to_ref = path_to_ref

        # self.sample_names_check()

        self.chr_starts, self.genome_length, self.scaf_names = helper.genomestats(self.path_to_ref)

    # def sample_names_check(self):

    #     counter = 0
    #     for i, sample_name in enumerate(self.sample_names):

    #         while counter <= 3:
    #             if not (sample_name in str(self.diversity_files[i])):
    #                 print(f"Warning! Sample name {sample_name} not found in the name of corresponding diversity file {self.diversity_files[i]}. Continuing...")
    #                 counter += 1

    #         if counter > 3:
    #             print("Not printing any further warnings...")
            
    def main(self):

        # Get sample names
        print('Processing sample names...')
        with open(self.path_to_sample_names_file, 'r') as f:
            self.sample_names = f.read().splitlines()
        nsamples = len(self.sample_names)  # save number of samples
        print('Total number of samples: ' + str(nsamples))


        # Get all candidate SNP positions
        print('Processing candidate SNP positions...')
        combine = CombinePositions(self.path_to_diversity_files, self.path_to_ref)
        self.p = combine.main()


        dim=8

        self.counts = np.zeros((dim, len(self.p), nsamples), dtype='uint')  # initialize
        self.indel_counter = np.zeros((2, len(self.p), nsamples), dtype='uint')
        self.quals = np.zeros((len(self.p), nsamples), dtype='int')

        print('Processing diversity data...')
        for i, diversity_file in enumerate(self.diversity_files):

            with gzip.open(diversity_file, 'rb') as f:
                diversity = pickle.load(f)

                data = diversity['data']
                quals_sample = diversity['quals'].flatten()

            self.counts[:, :, i] = data[self.p - 1, 0:dim].T  # -1 convert position to index
            self.indel_counter[:, :, i] = data[self.p - 1, 8:10].T  # Num reads supporting indels and reads supporting deletions
            self.quals[:, i] = quals_sample[self.p - 1]  # -1 is to convert position to index

        # ## counts: counts for each base from forward and reverse reads at each candidate position for all samples
        # print('Gathering counts data at each candidate position...\n')

        # # Import list of directories for where to diversity file for each sample
        # dim=8
        # self.counts = np.zeros((dim, len(self.p), nsamples), dtype='uint')  # initialize
        # self.indel_counter = np.zeros((2, len(self.p), nsamples), dtype='uint')

        # for i, mpileup_file in enumerate(self.path_to_mpileup_files):

        #     data, coverage = Pileup2Diversity.pileup2diversity(mpileup_file,
        #                                                        self.chr_starts,
        #                                                        self.genome_length,
        #                                                        self.scaf_names)

        #     self.counts[:, :, i] = data[self.p - 1, 0:dim].T  # -1 convert position to index
        #     self.indel_counter[:, :, i] = data[self.p - 1, 8:10].T  # Num reads supporting indels and reads supporting deletions


        # ## Quals: quality score (relating to sample purity) at each position for all samples
        # print('Gathering quality scores at each candidate position...')
        # # Import list of directories for where to quals for each sample

        # self.quals = np.zeros((len(self.p), nsamples), dtype='int')  # initialize

        # for i, vcf_file in enumerate(self.path_to_vcf_files):
            
        #     quals_sample = Vcf2Quals.vcf_to_quals(vcf_file,
        #                                           self.chr_starts,
        #                                           self.genome_length,
        #                                           self.scaf_names)
                    

        #     quals_sample = quals_sample.flatten()
        #     self.quals[:, i] = quals_sample[self.p - 1]  # -1 is to convert position to index


        self.write_CMT()

    def write_CMT(self):

        outdir = os.path.dirname(self.path_to_out_cmt)
        if not os.path.exists(outdir):
            os.makedirs(outdir)

        CMT = {'sample_names': self.sample_names,
               'p': self.p,
               'counts': self.counts,
               'quals': self.quals,
               'indel_counter': self.indel_counter}

        file_path = self.path_to_out_cmt
        print("Saving " + self.path_to_out_cmt)
        with gzip.open(file_path, 'wb') as f:
            pickle.dump(CMT, f)

    # @staticmethod
    # def parse_file_list(path_to_file):
    #     with open(path_to_file, 'r') as f:
    #         file_list = f.read().splitlines()

    #     for file_ in file_list:
    #         if not os.path.isabs(file_):
    #             abs_path = os.path.abspath(file_)
    #             if not os.path.exists(abs_path):
    #                 raise FileNotFoundError(f"File not found: {abs_path}")
    #             file_list[file_list.index(file_)] = os.path.abspath(file_)

    #     return file_list



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Module containing scripts for StrainSlicer Snakemake
@author: evanqu
"""
from Bio import SeqIO
import os
import numpy as np
import pickle
import glob
import subprocess
import gzip

# modified read_move_link_samplesCSV.py         
def read_samples_CSV_classify(spls):
	# reads in samples_case.csv file, format: Path,Sample,ReferenceGenome,Outgroup
	hdr_check = ['Path','Sample','FileName','Classifier','ReferenceGenome','Outgroup']
	switch = "on"
	file = open(spls, 'r')
    #initialize lists
	list_path = []; list_splID = []; list_fileN = []
	list_cfrN = []; list_refG = []; list_outgroup = []
	for line in file:
		line = line.strip('\n').split(',')
		# Test Header. Note: Even when header wrong code continues (w/ warning), but first line not read.
		if switch == "on":
			if (line == hdr_check):
				print("Passed CSV header check")
			else:
				Warning("CSV did NOT pass header check! Code continues, but first line ignored")
			switch = "off"
			continue
		# build lists
		list_path.append(line[0])
		list_splID.append(line[1])
		list_fileN.append(line[2])
		list_cfrN.append(line[3])
		list_refG.append(line[4])
		list_outgroup.append(line[5])
	return [list_path,list_splID,list_fileN,list_cfrN,list_refG,list_outgroup]

def read_samples_CSV_makeclassifier(spls):
	hdr_check = ['Path','Sample','FileName','ReferenceGenome','Outgroup']
	switch = "on"
	file = open(spls, 'r')
    #initialize lists
	list_path = []; list_splID = []; list_fileN = []; list_refG = []; list_outgroup=[]
	for line in file:
		line = line.strip('\n').split(',')
		# Test Header. Note: Even when header wrong code continues (w/ warning), but first line not read.
		if switch == "on":
			if (line == hdr_check):
				print("Passed CSV header check")
			else:
				Warning("CSV did NOT pass header check! Code continues, but first line ignored")
			switch = "off"
			continue
		# build lists
		list_path.append(line[0])
		list_splID.append(line[1])
		list_fileN.append(line[2])
		list_refG.append(line[3]) 
		list_outgroup.append(line[4])
	return [list_path,list_splID,list_fileN,list_refG,list_outgroup]


def split_samplesCSV_classify(PATH_ls , SAMPLE_ls , FILENAME_ls, CLASSIFIER_ls , REF_Genome_ls):
    # Evan custom for strainslicer
    # takes info extracted form samples.csv; saves each line of samples.csv as sample_info.csv in data/{sampleID}
    for i, sample in enumerate(SAMPLE_ls):
        # get info for this sample
        sample_info_csv_text = PATH_ls[i] + ',' + SAMPLE_ls[i] + ',' + FILENAME_ls[i] + ',' + CLASSIFIER_ls[i] + ',' + REF_Genome_ls[i]
        #print( sample )
        #print( sample_info_csv_text )
        # make data directory for this sample if it doesn't already exist
        if not(os.path.isdir('data/' + sample)):
            os.makedirs('data/' + sample, exist_ok=True)
        # check to see if this mini csv with sample info already exists
        if os.path.isfile('data/' + sample + '/sample_info.csv'):
            # if so, read file
            old_file = open('data/' + sample + '/sample_info.csv','r')
            old_info = old_file.readline()
            old_file.close()
            # check to see if the existing file is consistent with samples.csv
            if not(old_info == sample_info_csv_text):
                # if not, remove the old file and save sample info in a new file
                #print('Information file must be updated.')  
                os.remove('data/' + sample + '/sample_info.csv')
                f = open('data/' + sample + '/sample_info.csv','w')
                f.write(sample_info_csv_text) 
                f.close()
            #else:
            #print('Information file already updated.')              
        else: # if mini csv with sample info does not already exist
            # save sample info in mini csv
            #print('Information file must be created.')  
            f = open('data/' + sample + '/sample_info.csv','w')
            f.write(sample_info_csv_text) 
            f.close()
            
def split_samplesCSV_makeclassifier(PATH_ls,SAMPLE_ls,REF_Genome_ls,FILENAME_ls,OUTGROUP_ls):
    ''' For make_classifier: takes info from samples.csv & saves each line as sample_info.csv in data/{sampleID}'''
    
    for i, sample in enumerate(SAMPLE_ls):
        # get info for this sample
        sample_info_csv_text = f"{PATH_ls[i]},{SAMPLE_ls[i]},{REF_Genome_ls[i]},{FILENAME_ls[i]},{OUTGROUP_ls[i]}"

        # make data directory for this sample if it doesn't already exist
        if not(os.path.isdir('data/' + sample)):
            os.makedirs('data/' + sample, exist_ok=True)
        # check to see if this mini csv with sample info already exists
        if os.path.isfile('data/' + sample + '/sample_info.csv'):
            # if so, read file
            old_file = open('data/' + sample + '/sample_info.csv','r')
            old_info = old_file.readline()
            old_file.close()
            # check to see if the existing file is consistent with samples.csv
            if not(old_info == sample_info_csv_text):
                # if not, remove the old file and save sample info in a new file
                #print('Information file must be updated.')
                os.remove('data/' + sample + '/sample_info.csv')
                f = open('data/' + sample + '/sample_info.csv','w')
                f.write(sample_info_csv_text)
                f.close()
            #else:
            #print('Information file already updated.')
        else: # if mini csv with sample info does not already exist
            # save sample info in mini csv
            #print('Information file must be created.')
            f = open('data/' + sample + '/sample_info.csv','w')
            f.write(sample_info_csv_text)
            f.close()

            
def findfastqfile(dr,ID,filename):
    fwd=[]
    rev=[]
    #search for filename as directory first
    potentialhits_forward=glob.glob(dr + '/' + filename +'/*1.fastq.gz')
    potentialhits_reverse=glob.glob(dr + '/' + filename +'/*2.fastq.gz')
    if len(potentialhits_forward)==1 and len(potentialhits_reverse)==1:
        fwd=potentialhits_forward[0]
        rev=potentialhits_reverse[0]
    #then search for filename as file.gz
    elif len(potentialhits_forward)==0 and len(potentialhits_reverse)==0:
        potentialhits_forward=glob.glob(dr + '/' + filename +'*1.fastq.gz')
        potentialhits_reverse=glob.glob(dr + '/' + filename +'*2.fastq.gz')
        if len(potentialhits_forward)==1 and len(potentialhits_reverse)==1:
            fwd=potentialhits_forward[0]
            rev=potentialhits_reverse[0]
        #then search as unzipped file
        elif len(potentialhits_forward)==0 and len(potentialhits_reverse)==0:
            potentialhits_forward=glob.glob(dr + '/' + filename +'*1.fastq')
            potentialhits_reverse=glob.glob(dr + '/' + filename +'*2.fastq')
            if len(potentialhits_forward)==1 and len(potentialhits_reverse)==1:
                subprocess.run("gzip " + potentialhits_forward[0], shell=True)  
                subprocess.run("gzip " + potentialhits_reverse[0], shell=True)
                fwd=potentialhits_forward[0]+'.gz'
                rev=potentialhits_reverse[0]+'.gz'
            else:
                foldername=glob.glob(dr + '/' + filename + '*')
                if foldername and os.path.isdir(foldername[0]):
                    foldername=foldername[0]
                    potentialhits_forward=glob.glob(foldername + '/*' + filename + '*1*.fastq.gz')
                    potentialhits_reverse=glob.glob(foldername + '/*' + filename + '*2*.fastq.gz')
                    if len(potentialhits_forward)==1 and len(potentialhits_reverse)==1:
                        fwd=potentialhits_forward[0]
                        rev=potentialhits_reverse[0]
                    elif len(potentialhits_forward)==0 and len(potentialhits_reverse)==0:
                        print(foldername + '/*' + filename + '*2*.fastq.gz')
                        potentialhits_forward=glob.glob(foldername +  '/*' + filename + '*1*.fastq')
                        potentialhits_reverse=glob.glob(foldername + '/*' + filename + '*2*.fastq')
                        if len(potentialhits_forward)==1 and len(potentialhits_reverse)==1:
                            subprocess.run("gzip " + potentialhits_forward[0], shell=True)  
                            subprocess.run("gzip " + potentialhits_reverse[0], shell=True)
                            fwd=potentialhits_forward[0]+'.gz'
                            rev=potentialhits_reverse[0]+'.gz'
    if not(fwd) or not(rev):
        raise ValueError('Either no file or more than 1 file found in ' + dr + 'for ' + ID)
    ##zip fastq files if they aren't already zipped
    subprocess.run("gzip " + fwd, shell=True)   
    subprocess.run("gzip " + rev, shell=True)   
    return [fwd, rev]
    
def makelink(path,sample,filename):
    #When sample is run on a single lane
    #File name can be either a COMPLETE directory name or a file name in batch(called path in this fx)
    [fwd_file, rev_file]=findfastqfile(path,sample, filename)
    subprocess.run('ln -s -T ' + fwd_file + ' data/' + sample + '/R1.fq.gz', shell=True)    
    subprocess.run('ln -s -T ' + rev_file + ' data/' + sample + '/R2.fq.gz', shell=True)    


def cp_append_files(paths,sample,filename):
    #When sample is run on multiple lanes with same barcode
    fwd_list=''
    rev_list=''
    for path in paths:
        #Provider name can be either a COMPLETE directory name or a file name in batch(called path in this fx)
        [fwd_file, rev_file]=findfastqfile(path,sample, filename)
        fwd_list=fwd_list+ ' ' + fwd_file
        rev_list=rev_list+ ' ' + rev_file
        print(rev_list)
        print(fwd_list)
    subprocess.run("zcat " + fwd_list + ' | gzip > data/' +  sample + '/R1.fq.gz', shell=True)
    subprocess.run("zcat " + rev_list + ' | gzip > data/' +  sample + '/R2.fq.gz', shell=True)

def read_fasta(REFGENOME_DIR): 
    '''Reads in fasta file. If directory is given, reads in dir/genome.fasta
    Args:
        REFGENOME_DIR (str): Path to reference genome.

    Returns: SeqIO object for reference genome.
    '''
    fasta_file = glob.glob(REFGENOME_DIR + '/genome.fasta')
    if len(fasta_file) != 1:
        fasta_file_gz = glob.glob(REFGENOME_DIR + '/genome.fasta.gz')
        if len(fasta_file_gz) != 1:
            raise ValueError('Either no genome.fasta(.gz) or more than 1 genome.fasta(.gz) file found in ' + REFGENOME_DIR)
        else: # genome.fasta.gz
            refgenome = SeqIO.parse(gzip.open(fasta_file_gz[0], "rt"),'fasta')
    else: # genome.fasta
        refgenome = SeqIO.parse(fasta_file[0],'fasta')
    
    return refgenome

def genomestats(REFGENOMEFOLDER):
    '''Parse genome to extract relevant stats

    Args:
        REFGENOMEFOLDER (str): Directory containing reference genome file.

    Returns:
        ChrStarts (arr): DESCRIPTION.
        Genomelength (arr): DESCRIPTION.
        ScafNames (arr): DESCRIPTION.

    '''

    refgenome = read_fasta(REFGENOMEFOLDER)
    
    Genomelength = 0
    ChrStarts = []
    ScafNames = []
    for record in refgenome:
        ChrStarts.append(Genomelength) # chr1 starts at 0 in analysis.m
        Genomelength = Genomelength + len(record)
        ScafNames.append(record.id)
    # close file
    #refgenome.close() # biopy update SeqIO has no close attribute anymore.
    # turn to np.arrys!
    ChrStarts = np.asarray(ChrStarts,dtype=int)
    Genomelength = np.asarray(Genomelength,dtype=int)
    ScafNames = np.asarray(ScafNames,dtype=object)
    return ChrStarts,Genomelength,ScafNames

def p2chrpos(p, ChrStarts):
    '''Convert 1col list of pos to 2col array with chromosome and pos on chromosome

    Args:
        p (TYPE): DESCRIPTION.
        ChrStarts (TYPE): DESCRIPTION.

    Returns:
        chrpos (TYPE): DESCRIPTION.

    '''
        
    # get chr and pos-on-chr
    chromo = np.ones(len(p),dtype=int)
    if len(ChrStarts) > 1:
        for i in ChrStarts[1:]:
            chromo = chromo + (p > i) # when (p > i) evaluates 'true' lead to plus 1 in summation. > bcs ChrStarts start with 0...genomestats()
        positions = p - ChrStarts[chromo-1] # [chr-1] -1 due to 0based index
        chrpos = np.column_stack((chromo,positions))
    else:
        chrpos = np.column_stack((chromo,p))
    return chrpos

def combine_classifier_positions(cfrs,output_allpos_file,output_chrpos_file,refgenome_folder):
    
    all_pos = np.array([], dtype=np.int32)
    chr_starts, genome_length, scaf_names = genomestats(refgenome_folder)
    
    if len(cfrs) == 1:
        for c in list(cfrs):
            for filename in os.listdir(c):
                if filename.endswith('.classifier'):
                    with open(c+'/'+filename,'rb') as f:
                        csSNPs = pickle.load(f)
                    if len(csSNPs) != 3:
                        raise Exception('csSNP object is not correct shape!')
                    f_pos = csSNPs[1]
                    all_pos = np.unique(np.concatenate([f_pos,all_pos]))
                    
                else:
                    print('Warning! File '+filename+' does not have .classifier ending')
    if len(cfrs) > 1:
        print("Sorry, don't support multiple classifiers per run yet")

    all_pos.sort()
    print(str(len(all_pos)) +' total positions found across ' + str(len(os.listdir(list(cfrs)[0]))) + ' classifier(s)')
    
    chr_pos = p2chrpos(all_pos,chr_starts)
    chr_names = np.array([scaf_names[i-1] for i in chr_pos[:,0]])
    chr_pos_final = np.vstack((chr_names,chr_pos[:,1])).T
    #save as lists
    np.savetxt(output_allpos_file, all_pos, fmt='%i')
    np.savetxt(output_chrpos_file, chr_pos_final, delimiter='\t', fmt='%s')
    
    return


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb 18 14:50:42 2022

@author: evanqu
"""
import argparse
import os
import subprocess
import pickle
import gzip
import time
import datetime
import glob
import numpy as np
from Bio import AlignIO
from Bio import SeqIO

#%% Testing

# NOTE! THIS VERSION IS EXCLUSIVELY FOR TESTING. PULL FROM THE VERSION IN /phlame/snakemake_make_classifier/scripts/

os.chdir("/Users/evanqu/Dropbox (MIT)/Lieberman Lab/Personal lab notebooks/Evan/1-Projects/phlame_project/results/2022_03_makeclassifiers/Cacnes_acera_isolates_NEW2023")
input_cmt_file="Cacnes_repCMT.pickle.gz"
output_phylip="Cacnes_acera"
output_name_ids="Cacnes_acera_phylip2names.txt"
refGenome_file="/Users/evanqu/Dropbox (MIT)/Lieberman Lab/Personal lab notebooks/Evan/reference_genomes/Pacnes_C1.fasta"

min_cov_to_include=8
min_maf_for_call=0.85
min_strand_cov_for_call=2
min_qual_for_call=-30
min_presence_core=0.95
min_median_cov_samples=3
remov_recomb=True
consider_indels=False

# existing tree
# os.chdir("/Users/evanqu/Dropbox (MIT)/Lieberman Lab/Personal lab notebooks/Evan/1-Projects/phlame_project/results/2022_03_makeclassifiers/Sepi_acera_isolates_NEW2023")
# intree="tree/Sepi_acera_pars_norecomb.tre"
# phylip2names_file="tree/Sepi_acera_phylip2names.txt"
# outtree="tree/Sepi_acera_pars_norecomb_isonames.tre"


# cmt2phylip(input_cmt_file, output_phylip, output_name_ids, refGenome_file,
#             consider_indels=False, remov_recomb=False)

# rename_phylip(phylip2names_file, intree, outtree, outclustertree=False, rep_CMT_file=False)

np.savetxt('goodpos_60282.txt',pos[goodpos], fmt='%i')

#%%

def cmt2phylip(input_cmt_file, output_phylip, output_name_ids, refGenome_file,
               min_cov_to_include=8, min_maf_for_call=0.85,
               min_strand_cov_for_call=2, min_qual_for_call=-30,
               min_presence_core=0.987,min_median_cov_samples=3,
               consider_indels=False,remov_recomb=False):
    '''Finds fixed mutations within a set of samples and outputs phylip format file

    Args:
        input_cmt_file (str): Path to input candidate mutation table.
        output_phylip (str): Path to output phylip file.
        output_name_ids (str): Path to output file to rename phylip names to original.
        min_cov_to_include (float, optional): Minimum avg. coverage across positions to include a sample. Defaults to 8.
        min_maf_for_call (float, optional): Minimum major allele frequency to call a major allele for position. Defaults to 0.85.
        min_strand_cov_for_call (int, optional): Minimum coverage per strand to call a major allele for position. Defaults to 2.
        min_qual_for_call (int, optional): Minimum mapping quality to call a major allele for position. Defaults to -30.
        min_presence_core (float, optional): Minimum presence across samples to include a position. Defaults to 0.9.
        min_median_cov_samples (int, optional): Minimum median coverage across samples to include a position. Defaults to 3.
        consider_indels (bool, optional): Consider number of indels when filtering a position. Defaults to False.
        
    Returns:
        None.

    '''
    
    filterby_sample = {\
                       'min_cov_to_include':min_cov_to_include
                       }
    
    filterby_site_per_sample = {\
                              'min_maf_for_call':min_maf_for_call,
                              'min_strand_cov_for_call':min_strand_cov_for_call,
                              'min_qual_for_call': min_qual_for_call,
                              }
    
    filterby_site_across_samples = {\
                                  'min_presence_core':min_presence_core,
                                  'min_median_cov_samples':min_median_cov_samples,
                                  }
        
    filter_parameter_recombination = {\
                                'distance_for_nonsnp' : 300, #region in bp on either side of goodpos that is considered for recombination
                                'corr_threshold_recombination' : 0.75 #minimum threshold for correlation
                                }
    NTs = np.array(['A','T','C','G'],dtype=object) # NTs='ATCG'

        
    print("Reading in candidate mutation table....")
    
    sample_names, pos, counts, quals, indel_counter = read_counts_table(input_cmt_file)
    
    ### Modify some structures ###
    sample_names_all=np.asarray(sample_names,dtype=object)
    #reduce indel_counter to pxs matrix giving count for indels overall
    # indels_all=np.sum(indel_counter,axis=0)
    coverage_all=counts.sum(axis=0) #pxs
    
    print("Filtering samples...")
    
    #samples that pass sample-wise filters
    good_sample_bool = coverage_all.mean(axis=0) > float(filterby_sample['min_cov_to_include'])
    
    good_sample_names = sample_names_all[good_sample_bool]
    good_counts = counts[:,:,good_sample_bool]
    good_quals = quals[:,good_sample_bool]
    # good_indels = indels_all[:,good_sample_bool]
    
    num_samples=len(good_sample_names)
    coverage=good_counts.sum(axis=0)
    coverage_f_strand=good_counts[0:4,:,:].sum(axis=0) #should be pxs
    coverage_r_strand=good_counts[4:8,:,:].sum(axis=0)

    print(f"{num_samples}/{len(sample_names)} samples passed coverage filter.")
    print("The following samples did NOT pass coverage filter; NOT included:")
    print( sample_names_all[coverage_all.mean(axis=0) < float(filterby_sample['min_cov_to_include'])] )
    
    if num_samples < 3:
        raise Exception("Too few samples fullfill filter criteria!")
        
    print("Filtering positions...")
    #Get major allele at each position
    [maNT, maf, minorNT, minorAF] = get_major_allele_nt(good_counts)
    
    # Define mutations that pass filters in each and across samples.
    calls = np.copy(maNT)
    #remember quals are negative!
    calls[ good_quals > float(filterby_site_per_sample['min_qual_for_call']) ] = -1
    calls[ maf < float(filterby_site_per_sample['min_maf_for_call']) ] = -1
    calls[ coverage_f_strand < float(filterby_site_per_sample['min_strand_cov_for_call']) ] = -1
    calls[ coverage_r_strand < float(filterby_site_per_sample['min_strand_cov_for_call']) ] = -1
    
    if consider_indels:
        indels_all=np.sum(indel_counter,axis=0)
        good_indels = indels_all[:,good_sample_bool]
        calls[good_indels > (0.5*coverage)] = -1 #50%+ of reads @ site cannot support indel
        
    max_fracNs_bool = ( ((calls>-1).sum(axis=1)) <= \
                        (num_samples*float(filterby_site_across_samples['min_presence_core'])) )
    min_med_cov_bool = ( np.median(coverage, axis=1) < \
                         float(filterby_site_across_samples['min_median_cov_samples']) )
    
    site_filter_bool=np.any((max_fracNs_bool,min_med_cov_bool),axis=0)
    calls[site_filter_bool,:] = -1 # sites that fail qc ->-1, for all samples      
    
    print('Done filter.')
    
    print("Finding within-sample SNPs...")
    mut_qual,mut_qual_isolates = ana_mutation_quality(calls,good_quals)
    
    # nan in mutQual gives a warning, which is fine (nan=False=excluded)
    fixedmutation = (calls>-1)&(np.tile(mut_qual,(1,num_samples)) <= 1)
    hasmutation = np.any((fixedmutation == True,),axis=0)
    # NOTE: filteredpos is INDEX of filtered positions for p!
    filteredpos = np.where(np.sum(hasmutation, axis=1)> 0)[0] 
    
    # remove pos that are not variable between samples 
    # (fixed different to outgroup or N)
    bool_allN_outgroup = (calls>-1)
    # For outgroup functionality: NOT added yet!
    # bool_allN_outgroup[:,outgroup_idx] = False 
    
    #indices of nonvariable positions
    non_variable_idx = []
    for i,fp in enumerate(filteredpos):
        # identify sites non-variable between samples
        if len(np.unique(calls[fp,:][bool_allN_outgroup[fp,:]])) == 1:
            non_variable_idx.append(i)
    
    # NOTE: goodpos is INDEX of good positions for p!
    goodpos = np.delete(filteredpos,non_variable_idx)
    
    print(str(len(goodpos)) + ' goodpos found.')
    
    if len(goodpos) < 3:
        raise Exception("Too few positions after filter!")
        
    #  Check for recombination in p and remove positions from goodpos
    if remov_recomb:
        print("Filtering recombinant positions...")
        # When no outgroup defined: refnt ~= ancnt:
        [chrStarts, genomeLength, scafNames] = genomestats(refGenome_file);

        refnt = extract_outgroup_mutation_positions(refGenome_file, p2chrpos(pos,chrStarts));
        ancnt = refnt
        ancnti_m = np.full(ancnt.shape, 9)
        for idx,allele in enumerate(ancnt):
            if allele in NTs:
                ancnti_m[idx,] = np.where(NTs==allele)[0][0] # strip down to index number
            else:
                ancnti_m[idx,] = -1

        recombpos = findrecombinantSNPs(pos, goodpos, good_counts, ancnti_m, num_samples, 
                                        filter_parameter_recombination['distance_for_nonsnp'],
                                        filter_parameter_recombination['corr_threshold_recombination'])

        #These are the positions in p that are likely recombinant that we will remove from goodpos
        print(str(sum(np.isin(goodpos, recombpos))) + ' of a total ' + str(goodpos.shape[0]) + ' ('  + str(sum(np.isin(goodpos, recombpos))/goodpos.shape[0]*100) + '%) positions in goodpos were found to be recombinant.')
        goodpos=goodpos[~np.isin(goodpos, recombpos)]
    
    print("Writing phylip file...")
    # grab calls only at goodpos
    # numpy broadcasting of row_array requires np.ix_()
    calls_for_treei=calls[np.ix_(goodpos)]
    # Convert -10123 to NATCG translation
    calls_for_tree = idx2nts(calls_for_treei) 
    
    sample_names_4phylip = np.char.add(np.arange(0,len(good_sample_names)).astype(str), \
                                      good_sample_names.astype(str)).astype(object)
    
    #.dnapars.fa > for dnapars...deleted later
    write_calls_to_fasta(calls_for_tree,sample_names_4phylip,output_phylip+".dnapars.fa") 
    # turn fa to phylip and delete fasta with short tip labels    
    aln = AlignIO.read(output_phylip+".dnapars.fa", 'fasta')
    AlignIO.write(aln, output_phylip+".phylip", "phylip")
    subprocess.run(["rm -f " + output_phylip+".dnapars.fa"],shell=True)

    # Write object to convert phylip names back at the end
    with open(output_name_ids,'w') as f:
        for line in range(len(good_sample_names)):
            f.write(f"{sample_names_4phylip[line][:10]}\t{good_sample_names[line]}\n")

def read_counts_table(path_to_cmt_file):
    
    if path_to_cmt_file.endswith('.pickle.gz'):
        with gzip.open(path_to_cmt_file,'rb') as f:
            CMT = pickle.load(f)
            
            if type(CMT)==dict:
                counts = CMT['counts']; sample_names = CMT['sample_names']
                pos = CMT['p']; quals=CMT['quals']; indel_counter=CMT['indel_counter']
            else:
                counts = CMT[2]; sample_names=CMT[0]; pos = CMT[1]; quals = CMT[3]; indel_counter=CMT[4]
            
    elif path_to_cmt_file.endswith('.pickle'):
        with open(path_to_cmt_file,'rb') as f:
            CMT = pickle.load(f)
            
            if type(CMT)==dict:
                counts = CMT['counts']; sample_names = CMT['sample_names']
                pos = CMT['p']; quals=CMT['quals']; indel_counter=CMT['indel_counter']
            else:
                counts = CMT[2]; sample_names=CMT[0]; pos = CMT[1]; quals = CMT[3]; indel_counter=CMT[4]
    
    else:
        raise IOError("Error: candidate mutation table does not end in .pickle or .pickle.gz!")
        
    if np.size(counts,axis=0) != 8:
        print('Transposing counts table...')
        counts = counts.transpose(1,2,0)
        
    return [sample_names, pos, counts, quals, indel_counter]

def genomestats(REFGENOMEFILE):
    # parse ref genome to extract relevant stats
    refgenome = SeqIO.parse(REFGENOMEFILE,'fasta')
    Genomelength = 0
    ChrStarts = []
    ScafNames = []
    for record in refgenome:
        ChrStarts.append(Genomelength) # chr1 starts at 0 in analysis.m
        Genomelength = Genomelength + len(record)
        ScafNames.append(record.id)
    # turn to np.arrys!
    ChrStarts = np.asarray(ChrStarts,dtype=int)
    Genomelength = np.asarray(Genomelength,dtype=int)
    ScafNames = np.asarray(ScafNames,dtype=object)
    return [ChrStarts,Genomelength,ScafNames]

def get_major_allele_nt(counts):
    
    c=counts[0:4,:,:]+counts[4:8,:,:]; # flatten frw and rev ATCG counts    

    sorted_arr = np.sort(c,axis=0) #sort by ATCG counts
    sortedpositions = np.argsort(c,axis=0) # return matrix indices of sort
    
    # get allele counts for major allele (4th row)
    # weird "3:4:" indexing required to maintain 3d structure
    maxcount = sorted_arr[3:4:,:,:] 
    # get allele counts for first minor allele (3rd row)
    # tri/quadro-allelic ignored!!
    minorcount = sorted_arr[2:3:,:,:] 
    
    with np.errstate(divide='ignore', invalid='ignore'):
        maf = maxcount / sorted_arr.sum(axis=0,keepdims=True)
        minorAF = minorcount / sorted_arr.sum(axis=0,keepdims=True)
    maf = np.squeeze(maf,axis=0) # turn 2D; axis=1 to keep 2d structure when only one position!
    maf[np.isnan(maf)]=0 # set to 0 to indicate no data
    minorAF = np.squeeze(minorAF,axis=0) 
    minorAF[np.isnan(minorAF)]=0 # set to 0 to indicate no data/no minor AF
    
    # index position in sortedpositions represents allele position ATCG;
    # A=0,T=1,C=2,G=3
    # axis=1 to keep 2d structure when only one position!
    majorNT = np.squeeze(sortedpositions[3:4:,:,:],axis=0) 
    minorNT = np.squeeze(sortedpositions[2:3:,:,:],axis=0)

    # Note: If counts for all bases are zero, then sort won't change the order
    # (since there is nothing to sort), thus majorNT/minorNT will be put to -1 (NA)
    # using maf (REMEMBER: minorAF==0 is a value!)
    majorNT[maf==0]=-1
    minorNT[maf==0]=-1
    
    return majorNT, maf, minorNT, minorAF

def ana_mutation_quality(calls,quals):
    # This function calls mutations within the data itself, instead of wrt reference 
    # It takes as input the called nucleotides (calls) and the quality score (qual)
    # and outputs only the mutation call quality (mut_qual) calculated as 
    # max_i(min_j(Qual_i, Qual_j)) for i,j over all samples with different calls
    # It also outputs the samples giving this maxmin quality (mut_qual_isolates)
    
    # If there is no within data mutation in a given position (all nucleotides
    # are equal, but different from the reference), mut_qual returns 0 at that
    # position.
    
    [n_muts, n_strain] = calls.shape ;
    mut_qual = np.zeros((n_muts,1)) ; 
    mut_qual_isolates = np.zeros((n_muts,2)); 
    
    # generate template index array to sort out strains gave rise to reported FQ values
    s_template=np.zeros( (len(calls[0,:]),len(calls[0,:])) ,dtype=object)
    for i in range(s_template.shape[0]):
        for j in range(s_template.shape[1]):
            s_template[i,j] = str(i)+"_"+str(j)

    for k in range(n_muts):
        if len(np.unique(np.append(calls[k,:], 4))) <= 2: # if there is only one type of non-N (4) call, skip this location
            mut_qual[k] = np.nan ;
            mut_qual_isolates[k,:] = 0; 
        else:
            c = calls[k,:] ; c1 = np.tile(c,(c.shape[0],1)); c2 = c1.transpose() # extract all alleles for pos k and build 2d matrix and a transposed version to make pairwise comparison
            q = quals[k,:] ; q1 = np.tile(q,(q.shape[0],1)); q2 = q1.transpose() # -"-
            g = np.all((c1 != c2 , c1 != 4 , c2 != 4) ,axis=0 )  # no data ==4; boolean matrix identifying find pairs of samples where calls disagree (and are not N) at this position
            #positive_pos = find(g); # numpy has no find; only numpy where, which does not flatten 2d array that way
            # get mut_qual + logical index for where this occurred
            mut_qual[k] = np.max(np.minimum(q1[g],q2[g])) # np.max(np.minimum(q1[g],q2[g])) gives lower qual for each disagreeing pair of calls, we then find the best of these; NOTE: np.max > max value in array; np.maximum max element when comparing two arryas
            MutQualIndex = np.argmax(np.minimum(q1[g],q2[g])) # return index of first encountered maximum!
            # get strain ID of reorted pair (sample number)
            s = s_template
            strainPairIdx = s[g][MutQualIndex]
            mut_qual_isolates[k,:] = [strainPairIdx.split("_")[0], strainPairIdx.split("_")[1]]
            
    return [mut_qual,mut_qual_isolates]


def idx2nts(calls, missingdata="?"):
    # translate index array to array containing nucleotides
    # add 5th element --> no data! == index -1
    nucl = np.array([missingdata,'A','T','C','G'],dtype=object) 
    palette = [-1,0,1,2,3] # values present in index-array
    index = np.digitize(calls.ravel(), palette, right=True)
    
    return nucl[index].reshape(calls.shape)

def write_calls_to_fasta(calls,sample_names,output_file):
    
    fa_file = open(output_file, "w")
    
    for i,name in enumerate(sample_names):
        nucl_string = "".join(list(calls[:,i]))
        fa_file.write(">" + name + "\n" + nucl_string + "\n")
    
    fa_file.close()

def extract_outgroup_mutation_positions(REFGENOMEFILE,position):
    # extracts the ref nucleotide for every position. positions needs to be sorted by chr
    # CMTpy=True: if old matlab build_candidate_mutation.mat used, put flag False. p 1-based correction
    refgenome = SeqIO.parse(REFGENOMEFILE,'fasta')
    refnt = np.zeros(position.shape[0],dtype=object)
    pos_counter = 0
    chr_counter = 1
    for record in refgenome:
        poschr = position[ position[:,0]==chr_counter , 1]
        for sglpos in poschr:
            refnt[pos_counter] = str(record.seq)[sglpos] 

            pos_counter += 1
        chr_counter += 1
    return refnt

def p2chrpos(p, ChrStarts):
    '''# return 2col array with chr and pos on chr
    #p...continous, ignores chr
    #pos: like p, 0-based'''

    # get chr and pos-on-chr
    chr = np.ones(len(p),dtype=int)
    if len(ChrStarts) > 1:
        for i in ChrStarts[1:]:
            chr = chr + (p > i) # when (p > i) evaluates 'true' lead to plus 1 in summation. > bcs ChrStarts start with 0...genomestats()
        positions = p - ChrStarts[chr-1] # [chr-1] -1 due to 0based index
        pos = np.column_stack((chr,positions))
    else:
        pos = np.column_stack((chr,p))
    return pos


def findrecombinantSNPs(pos, goodpos, good_counts, ancnti_m, num_samples, 
                        distance_for_nonsnp, 
                        corr_threshold_recombination):

    ancnti_rep=np.tile(ancnti_m, (num_samples, 1)).T     #make new array that is ancti_m long x Nsample across
    [cmajorNT, cmajorAF, cminorNT, cminorAF] = get_major_allele_nt(good_counts)     #generates data structures of allele frequencies
    #return majorNT, maf, minorNT, minorAF
    cminorAF[np.isnan(cminorAF)]=0     #set nan values to 0

    #create mutantAF
    mutantAF=np.zeros((cmajorNT.shape))
    mutantAF[np.where(cmajorNT!=ancnti_rep)] = cmajorAF[np.where(cmajorNT!=ancnti_rep)]
    mutantAF[np.where(cminorNT!=ancnti_rep)] = mutantAF[np.where(cminorNT!=ancnti_rep)]+cminorAF[np.where(cminorNT!=ancnti_rep)]

    #look for recombination regions
    nonsnp = []
    for i in range(len(goodpos)):
        if (i % 5000) == 0:
            print(".")
        gp = pos[goodpos[i]]
        #find nearby snps
        if gp > distance_for_nonsnp:
            region = np.array(np.where((pos > gp - distance_for_nonsnp) & (pos < gp + distance_for_nonsnp)) ).flatten()
            if len(region)>1: 
                r = mutantAF[region,:]
                corrmatrix = np.corrcoef(r) 
                [a,b]=np.where(corrmatrix > corr_threshold_recombination)
                nonsnp.extend(list(region[a[np.where(a!=b)]]))
    
    nonsnp=np.unique(nonsnp)
    
    return nonsnp

def generate_dnapars_tree(path_to_phylip_file,path_to_output_tree,
                          path_to_renaming_file=False):
    
    # Write alignment file (as fasta)
    # calc NJ or Parsimonous tree or None
    # writeDnaparsAlignment==True for writing dnapars input for usage on cluster
    ts = time.time()
    timestamp = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d_%H-%M-%S')
    
    if not path_to_phylip_file.endswith('.phylip'):
        path_to_phylip_file += '.phylip'
        
    # Find dnapars executable; searches up to 5 directories back
    dnapars_path = glob.glob('dnapars')
    path_extension = "../"
    backstop = 0
    while len(dnapars_path) == 0 and backstop <= 5:
        dnapars_path = glob.glob(path_extension+'dnapars')
        path_extension = path_extension + "../"
        backstop = backstop + 1
    if len(dnapars_path) == 0:
        raise ValueError('Error: dnapars executable could not be located.')
    elif dnapars_path[0]=='dnapars':
        dnapars_path[0] = './dnapars'
    
    # Write parameters file
    with open(f"{path_to_output_tree}_{timestamp}.options.txt",'w') as file:
        file.write(path_to_phylip_file+"\n")
        file.write("f"+"\n")
        file.write(f"{path_to_output_tree}_{timestamp}.dnapars"+"\n")
        file.write("5"+"\n")
        file.write("V"+"\n")
        file.write("1"+"\n")
        file.write("y"+"\n")
        file.write("f"+"\n")
        file.write(f"{path_to_output_tree}_{timestamp}.tre"+"\n"+"\n") #Path to tree

    # Run dnapars
    print("Building parsimony tree...")
    print( f"{dnapars_path[0]} < {path_to_output_tree}_{timestamp}_options.txt > {path_to_output_tree}_{timestamp}_dnapars.log")
    subprocess.run([ "touch outtree"  ],shell=True)
    subprocess.run([ f"{dnapars_path[0]} < {path_to_output_tree}_{timestamp}.options.txt > {path_to_output_tree}_{timestamp}.dnapars.log"  ],shell=True)
    print("Done!")
    
    # Re-write tree with new long tip labels  
    if path_to_renaming_file:
        print("Renaming tree with original labels...")
        path_to_output_renamed_tree=f"{path_to_output_tree}_isonames.tre"
        rename_phylip(path_to_renaming_file,f"{path_to_output_tree}.tre",path_to_output_renamed_tree)

    return timestamp

def rename_phylip(phylip2names_file, intree, outtree, 
                  outclustertree=False, rep_CMT_file=False):
    '''Given a renaming file, rename 10chr phylip names into long format

    Args:
        phylip2names_file (TYPE): DESCRIPTION.
        intree (TYPE): DESCRIPTION.
        outtree (TYPE): DESCRIPTION.

    Returns:
        None.

    '''
    # Get phylip2names as dict
    phylip2names=dict()
    with open(phylip2names_file) as f:
        for line in f:
            key, value = line.strip().split('\t')
            phylip2names[key] = value
            
    # Replace phylip tree names
    with open(intree) as f:
        nwk=f.read()
    #Replace with representative isolate name
    for i in phylip2names.keys():
        nwk=nwk.replace(i,phylip2names[i])
    with open(outtree,'w') as f:
        f.write(nwk)
    
    if outclustertree: # Optionally output tree named by cluster
    
        # Get which tree isolate belongs to which cluster
        with gzip.open(rep_CMT_file,'rb') as f:
            CMT=pickle.load(f); sample_names=CMT[0]; cluster_IDs=CMT[4]
        tree2cluster=dict()
        for sam,clu in zip(sample_names,cluster_IDs):
            tree2cluster[sam]=clu
            
        #Replace with cluster name
        for i in tree2cluster.keys():
            nwk=nwk.replace(i,'Cluster '+tree2cluster[i])
        with open(outclustertree,'w') as f:
            f.write(nwk)
            
            
#%%
# import matplotlib.pyplot as plt

# num_samples=len(sample_names)

# # When no outgroup defined: refnt ~= ancnt:
# [chrStarts, genomeLength, scafNames] = genomestats(refGenome_file);

# refnt = extract_outgroup_mutation_positions(refGenome_file, p2chrpos(pos,chrStarts));
# ancnt = refnt
# ancnti_m = np.full(ancnt.shape, 9)
# for idx,allele in enumerate(ancnt):
#     if allele in NTs:
#         ancnti_m[idx,] = np.where(NTs==allele)[0][0] # strip down to index number
#     else:
#         ancnti_m[idx,] = -1
        

# ancnti_rep=np.tile(ancnti_m, (num_samples, 1)).T     #make new array that is ancti_m long x Nsample across
# [cmajorNT, cmajorAF, cminorNT, cminorAF] = get_major_allele_nt(counts)     #generates data structures of allele frequencies
# #return majorNT, maf, minorNT, minorAF
# cminorAF[np.isnan(cminorAF)]=0     #set nan values to 0

# #create mutantAF
# mutantAF=np.zeros((cmajorNT.shape))
# mutantAF[np.where(cmajorNT!=ancnti_rep)] = cmajorAF[np.where(cmajorNT!=ancnti_rep)]
# mutantAF[np.where(cminorNT!=ancnti_rep)] = mutantAF[np.where(cminorNT!=ancnti_rep)]+cminorAF[np.where(cminorNT!=ancnti_rep)]

# distance_for_nonsnp=500
# corr_threshold_recombination=0.75
# goodpos=np.arange(0,len(pos))

# #look for recombination regions
# nonsnp = []
# for i in range(len(goodpos)):
#     gp = pos[goodpos[i]]
#     #find nearby snps
#     if gp > distance_for_nonsnp:
#         region = np.array(np.where((pos > gp - distance_for_nonsnp) & (pos < gp + distance_for_nonsnp)) ).flatten()
#         if len(region)>1: 
#             r = mutantAF[region,:]
#             corrmatrix = np.corrcoef(r)
            
            
#             [a,b]=np.where(corrmatrix > corr_threshold_recombination)
#             nonsnp.extend(list(region[a[np.where(a!=b)]]))

# nonsnp=np.unique(nonsnp)

# corrmatrix[np.isnan(corrmatrix)]=0
# fig,axs=plt.subplots()
# hmap = axs.imshow(corrmatrix, cmap='Greys', interpolation='nearest')
# fig.colorbar(hmap, ax=axs)
        
#%% Main

if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument('-i', dest='Input', type=str, help='Path to input candidate mutation table.',required=True)
    parser.add_argument('-p', dest='Phylip', type=str, help='Path to output phylip.', required=True)
    parser.add_argument('-o', dest='Output', type=str, help='Path to output tree.', required=True)
    parser.add_argument('-n', dest='NameIDs', type=str, help='Path to output renaming file.', required=True)
    parser.add_argument('-r', dest='RefGenome', type=str, help='Path to reference genome file.', required=True)
    parser.add_argument('--min_cov', dest='MinCov', type=str, help='Minimum average coverage across positions to include a sample. Default=8.',required=False, default=8)
    parser.add_argument('--min_maf', dest='MinMAF', type=str, help='Minimum major allele frequency to call a major allele. Default=0.85.', required=False, default=0.85)
    parser.add_argument('--min_strand_cov', dest='MinStrandCov', type=str, help='Minimum coverage per strand to call a major allele. Default=2.', required=False, default=2)
    parser.add_argument('--min_qual', dest='MinQual', type=str, help='Minimum mapping quality to call a major allele. Default=-30.',required=False,default=-30)
    parser.add_argument('--min_presence_core', dest='MinCore', type=str, help='Minimum presence across samples to include a position. Default=0.9.', required=False,default=0.9)
    parser.add_argument('--min_median_cov_samples', dest='MinCovSamples', type=str, help='Minimum median coverage across samples to include a position. Default=3.', required=False, default=3)
    parser.add_argument('--filter_indels', dest='FilterIndels', type=str, help='Filter positions based on number of indels. Defaults to False.', required=False, default=False)
    parser.add_argument('--remov_recomb', dest='Recomb', type=bool, help='Filter positions based on number of indels. Defaults to False.', required=False, default=False)


    args = parser.parse_args()

    if args.Recomb:
        print('Remove recombinant regions: Yes')
        
    cmt2phylip(args.Input, args.Phylip, args.NameIDs, args.RefGenome,
               min_cov_to_include=args.MinCov, min_maf_for_call=args.MinMAF,
               min_strand_cov_for_call=args.MinStrandCov, min_qual_for_call=args.MinQual,
               min_presence_core=args.MinCore,min_median_cov_samples=args.MinCovSamples,
               consider_indels=args.FilterIndels, remov_recomb=args.Recomb)

    generate_dnapars_tree(args.Phylip,args.Output,
                          path_to_renaming_file=args.NameIDs)



    

    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Helper functions and classes for phlame.

@author: evanqu
"""

import os
import pickle
import gzip
import numpy as np
import pandas as pd
from Bio import SeqIO
import glob

class Frequencies():
    '''
    Holds clade frequency information from a given sample.
    '''
    
    def __init__(self, path_to_frequencies_file):
        
        self.freqs = pd.read_csv(path_to_frequencies_file,
                                       index_col=0)
                
class FrequenciesData():
    '''
    Holds clade specific SNV counts and modeling information from a given sample.
    '''

    def __init__(self, path_to_data_file):
        
        with gzip.open(path_to_data_file, 'rb') as f:
            
            data_dct, fit_info_dct = pickle.load(f)
            
            # clade_counts structured as follows
            
            self.clade_counts = data_dct['clade_counts']
            self.clade_counts_pos = data_dct['clade_counts_pos']
            
            self.counts_MLE = fit_info_dct['counts_MLE']
            self.total_MLE = fit_info_dct['total_MLE']
            self.counts_MAP = fit_info_dct['counts_MAP']
            self.chain = fit_info_dct['chain']
            self.prob = fit_info_dct['prob']


# class CountsMat():
#     '''
#     Holds data and methods for a counts matrix.
#     '''
#     def __init__(self, path_to_cts_file):
        
#         with gzip.open(path_to_cts_file,'rb') as f:
#             counts, pos = pickle.load(f)
                
#         self.counts = counts
#         self.pos = pos


class CountsMat():
    '''
    Holds data and methods for a counts matrix.
    '''

    def __init__(self, 
                 path_to_pileup, 
                 path_to_ref, 
                 path_to_classifiers):
        
        self.path_to_pileup = path_to_pileup
        self.path_to_ref = path_to_ref
        self.path_to_classifiers = path_to_classifiers

        self.chr_starts, self.genome_length, self.scaf_names = genomestats(self.path_to_ref)

    def readin(self,
               path_to_cts_file):
        '''
        Read in an existing counts object.
        '''
        
        with gzip.open(path_to_cts_file,'rb') as f:
            counts, pos = pickle.load(f)
                
        self.counts = counts
        self.pos = pos

    def main(self):
        
        print(f"Reading pileup file: {self.path_to_pileup}")
        counts, pos = self.pileup2counts(self.path_to_pileup,
                                         self.path_to_classifiers)
        
        self.counts = counts
        self.pos = pos

    def pileup2counts(self,
                      input_pileup,
                      path_to_classifiers):
        '''Grabs relevant allele info from mpileupfile and stores as a nice array.

        Args:
            input_pileup (str): Path to input pileup file.
            path_to_ref (str): Path to reference genome file.
            path_to_classifiers (str): Path to classifier file(s).
            
        '''
        # Initial parameters
        nts = 'ATCGatcg'
        num_fields = 8
        
        # Get classifier position information
        allpos = self.classifier_stats(path_to_classifiers)
        
        data = np.zeros((len(allpos),num_fields)) #format [[A T C G  a t c g],[...]]
        
        ##### read in mpileup file #####
        mpileup = open(input_pileup)

        loading_bar = 0
        for line in mpileup:
            
            # loading_bar+=1
            # if loading_bar % 500 == 0:
            #     print('.')

            lineinfo = line.strip().split('\t')
            
            #holds info for each line before storing in data
            temp = np.zeros((num_fields))
            
            chromo = lineinfo[0]
            
            #position (absolute)
            if chromo not in self.scaf_names:
                raise ValueError("Contig name in pileup file not found in reference!")

            if len(self.chr_starts) == 1:
                position=int(lineinfo[1])
            else:
                position=int(self.chr_starts[np.where(chromo==self.scaf_names)]) + int(lineinfo[1])
                #chr_starts begins at 0
            pidx = np.searchsorted(allpos, position) # index of position on allpos
            
            #ref allele
            ref=int(np.char.find(nts,lineinfo[2])) # convert to 0123
            if ref > 4:
                ref = ref - 4
            
            #calls info
            #calls=lineinfo[4]
            calls=np.array([ord(l) for l in lineinfo[4]]) #ASCII
            
            #find starts of reads ('^' in mpileup)
            startsk=np.where(calls==94)[0]
            for k in startsk:
                calls[k:k+2]=-1 #WHAT IS -1
                #remove mapping character, absolutely required because the next chracter could be $
            
            #find ends of reads ('$' in mpileup)
            endsk=np.where(calls==36)[0]
            calls[endsk]=-1
            
            #find indels + calls from reads supporting indels ('+-')
            indelk = np.where((calls==43) | (calls==45))[0]
            for k in indelk:
                if (calls[k+2] >=48) and (calls[k+2] < 58): #2 digit indel (size > 9 and < 100)
                    indelsize=int(chr(calls[k+1]) + chr(calls[k+2])) 
                    #indelsize=str2double(char(calls(k+1:k+2))); MATLAB
                    indeld=2
                else: #1 digit indel (size <= 9)
                    indelsize=int(chr(calls[k+1]))
                    indeld=1
            #remove indel info from counting
                calls[k:(k+1+indeld+indelsize)] = -1 #don't remove base that precedes an indel
            
            #replace reference matches (.,) with their actual calls
            if ref >=0:
                calls[np.where(calls==46)[0]]=ord(nts[ref]) #'.'
                calls[np.where(calls==44)[0]]=ord(nts[ref+4]) #','
            # if ref >=0:
            #     calls[np.where(calls==46)[0]]=ord(nts[int(np.char.find(nts,ref))]) # changed from nts(ref); matlab
            #     calls[np.where(calls==44)[0]]=ord(nts[int(np.char.find(nts,ref))+4]) # changed from=nts(ref+4); matlab

            #index reads for finding scores
            simplecalls=calls[np.where(calls>0)[0]]
            #simplecalls is a tform of calls where each calls position
            #corresponds to its position in bq, mq, td
            
            #count how many of each nt and average scores
            for nt in range(8):
                nt_count=np.count_nonzero(simplecalls == ord(nts[nt]))
                if nt_count > 0:
                    temp[nt]=nt_count
            # for nt in range(8):
            #     if not sum(simplecalls == ord(nts[nt])) == 0:
            #         temp[nt]=sum(simplecalls == ord(nts[nt]))

            # Store in big array
            data[pidx]=temp
        
        return data, allpos

    @staticmethod
    def read_fasta(path_to_refgenome): 
        '''Reads in fasta file. If directory is given, reads in dir/genome.fasta
        Args:
            path_to_refgenome (str): Path to reference genome.

        Returns: SeqIO object for reference genome.
        '''
        fasta_file = glob.glob(path_to_refgenome + '/genome.fasta')
        if len(fasta_file) != 1:
            fasta_file_gz = glob.glob(path_to_refgenome + '/genome.fasta.gz')
            if len(fasta_file_gz) != 1:
                raise ValueError('Either no genome.fasta(.gz) or more than 1 genome.fasta(.gz) file found in ' + path_to_refgenome)
            else: # genome.fasta.gz
                refgenome = SeqIO.parse(gzip.open(fasta_file_gz[0], "rt"),'fasta')
        else: # genome.fasta
            refgenome = SeqIO.parse(fasta_file[0],'fasta')
        
        return refgenome
    
    @staticmethod
    def classifier_stats(path_to_classifiers):
        '''Parse classifier file(s) to extract position information

        Args:
            path_to_classifiers (str): Comma separated string of paths to classifiers.

        Returns:
            allpos (arr): Array of absolute positions referenced in classifier(s).

        '''
        
        cat_pos = np.array([], dtype=np.int32)
    
        path_to_cfrs_ls = []
        # Parse whether file or directory of files
        if os.path.isdir(path_to_classifiers):
            
            for filename in os.listdir(path_to_classifiers):
                
                if filename.endswith('.classifier'):
                    path_to_cfrs_ls.append(path_to_classifiers+'/'+filename)
        else:
            path_to_cfrs_ls.append(path_to_classifiers)

        for cfr in path_to_cfrs_ls:
            
            with gzip.open(cfr,'rb') as f:
                csSNPs = pickle.load(f)
                                
                pos = csSNPs['cssnp_pos']
                cat_pos = np.concatenate([pos,cat_pos])

        allpos = np.unique(np.sort(cat_pos))
    
        return allpos

    @staticmethod        
    def chrpos_stats(self, path_to_pos_file):
        
        self.genome_length = 0
        self.scaf_names_ls = []
        chr_pos = []

        with open(path_to_pos_file,'r') as file:
            
            for line in file:
                
                position = line.strip().split('\t')
                self.scaf_names_ls.append(position[0])
                chr_pos.append(position[1])
        
        self.scaf_names, lengths = np.unique(self.scaf_names_ls,return_counts=True)
        self.chr_starts=[]

        if len(self.scaf_names) == 1:
            self.chr_starts.append(self.genome_length)
            self.genome_length = lengths[0]
        else:
            for scaf, len_ in zip(self.scaf_names,lengths):
                self.chr_starts.append(self.genome_length)            
                self.genome_length = self.genome_length + len_
        
        return self.chr_starts, self.genome_length, self.scaf_names


class PhlameClassifier():
    '''
    Holds data and methods for a single Phlame Classifier object
    '''
    def __init__(self,
                 csSNPs, csSNP_pos,
                 clades, clade_names):

        self.csSNPs = csSNPs
        self.csSNP_pos = csSNP_pos
        self.clades = clades
        self.clade_names = clade_names
        
        # Get allele information
        self.get_alleles()
    
    def read_file(path_to_classifier_file):

        with gzip.open(path_to_classifier_file, 'rb') as f:
            cssnp_dct = pickle.load(f)
            
            csSNPs = cssnp_dct['cssnps']
            csSNP_pos = cssnp_dct['cssnp_pos']
            clades = cssnp_dct['clades']
            clade_names = cssnp_dct['clade_names']
            
        return PhlameClassifier(csSNPs, csSNP_pos, clades, clade_names)

    def grab_level(self, PhyloLevel):
        '''
        Grab just information for a specific level.
        '''
        
        idx=[]
        
        for clade in PhyloLevel.clade_names:
            
            idx.append(np.where(self.clade_names==clade)[0][0])
        
        level_csSNPs = self.csSNPs[:,idx]

        level_csSNP_pos = self.csSNP_pos[~np.all(level_csSNPs == 0, axis=1)]
        
        return PhlameClassifier(level_csSNPs[~np.all(level_csSNPs == 0, axis=1)],
                                level_csSNP_pos,
                                PhyloLevel.clades,
                                PhyloLevel.names)
    
    def get_alleles(self):
        '''
        Get 1D list of every allele and corresponding clade.
        '''
        self.alleles = self.csSNPs[np.nonzero(self.csSNPs)]
        # corresponding clade index
        self.allele_cidx = np.nonzero(self.csSNPs)[1]


class Phylip():
    '''
    Holds data and methods for a phylip object.
    '''

    def check_valid(phylip_file):

        if not os.path.exists(phylip_file):
            return False
        
        with open(phylip_file, 'r') as f:
            line1 = f.readline().strip().split(' ')

        nsamples = int(line1[0])
        npos = int(line1[1])

        valid_samples = nsamples > 0
        valid_positions = npos > 0

        return valid_samples and valid_positions

class CandidateMutationTable():

    def __init__(self, path_to_cmt_file):
        '''
        Read in candidate mutation table from pickled object file.
        '''
        
        if path_to_cmt_file.endswith('.pickle.gz'):
            with gzip.open(path_to_cmt_file,'rb') as f:
                CMT = pickle.load(f)
        
        elif path_to_cmt_file.endswith('.pickle'):
            with open(path_to_cmt_file,'rb') as f:
                CMT = pickle.load(f)

        self.sample_names = np.array(CMT['sample_names'])
        self.counts = CMT['counts']
        self.pos = CMT['p']
        self.quals=CMT['quals']
        self.indel_counter=CMT['indel_counter']

        # FIX THIS!!!
        self.indel_counter = np.zeros((2,len(self.pos),
                                        len(self.sample_names)))
            
        # Note that 1 -> yes outgroup, 0 -> not outgroup
        # Makes booleans more confusing I know
        if 'in_outgroup' in CMT.keys():
            self.in_outgroup=CMT['in_outgroup']
            if type(self.in_outgroup[0]) == np.ndarray:
                self.in_outgroup = np.array(self.in_outgroup[0][0].split(' ')).astype(bool)

        else:
            self.in_outgroup = np.array([False]*len(self.sample_names))

        # Calculate coverage and indels
        self.calc_coverage()
        self.calc_indels_all()

    def calc_coverage(self):
        '''
        Calculate coverage for each sample.
        '''
        self.coverage = np.sum(self.counts,axis=0)

    def calc_indels_all(self):

        self.indels_all = np.sum(self.indel_counter,axis=0)

    # if np.size(self.counts, axis=0) != 8:
    #     print('Transposing counts table...')
    #     self.counts = self.counts.transpose(1,2,0)
            

def read_clades_file(path_to_clades_file, uncl_marker):
    '''
    Read in a clades file.
    '''        
    
    if not os.path.exists(path_to_clades_file):
        raise FileNotFoundError(f'File {path_to_clades_file} not found.')
    # Get delimiter
    with open(path_to_clades_file,'r') as file:
        firstline = file.readline()
   
    if len(firstline.strip().split('\t'))==2:
        dlim='\t'
    elif len(firstline.strip().split(','))==2:
        dlim=','
    else:
        raise ValueError('Delimiter in clades file not recognized. Please specify clades as a tab or comma separated list.')
    
    # Read in file
    clade_ids = np.loadtxt(path_to_clades_file, 
                           delimiter=dlim, 
                           dtype=str)
    
    # Reshape into dictionary
    clades_dct = dict()
    clade_names = []
    # Loop through unique clades & grab samples
    for clade in np.unique(clade_ids[:,1]):
        
        if clade==uncl_marker:
            continue
        
        isclade_bool = np.in1d(clade_ids[:,1], clade)
        clade_samples = clade_ids[isclade_bool,0].tolist()
        
        clades_dct[clade] = clade_samples
        clade_names.append(clade)
            
    return clades_dct, np.array(clade_names)

def rphylip(sample_names):
    '''Change : to | for consistency with phylip format'''
    
    rename = [sam.replace(':','|') for sam in sample_names]
    
    return np.array(rename)

def genomestats(path_to_refgenome_file):
    '''Extract relevant stats from a reference genome file.

    Args:
        path_to_refgenome_file (str): Path to reference genome file (.fasta).

    Returns:
        ChrStarts (TYPE): DESCRIPTION.
        Genomelength (TYPE): DESCRIPTION.
        ScafNames (TYPE): DESCRIPTION.

    '''
    
    refgenome = SeqIO.parse(path_to_refgenome_file,'fasta')
    
    Genomelength = 0
    ChrStarts = []
    ScafNames = []
    
    for record in refgenome:
        ChrStarts.append(Genomelength) # chr1 starts at 0 in analysis.m
        Genomelength = Genomelength + len(record)
        ScafNames.append(record.id)
    
    # turn to np.arrys
    ChrStarts = np.asarray(ChrStarts,dtype=int)
    Genomelength = np.asarray(Genomelength,dtype=int)
    ScafNames = np.asarray(ScafNames,dtype=object)
    
    return ChrStarts,Genomelength,ScafNames


def mant(counts):
    '''Get major and first minor allele along with frequencies for each position in a counts matrix. 

    Args:
        counts (arr): numpy-compatible array (8xpxs).

    Returns:
        maNT (TYPE): DESCRIPTION.
        maf (TYPE): DESCRIPTION.
        minorNT (TYPE): DESCRIPTION.
        minorAF (TYPE): DESCRIPTION.

    '''
    
    c=counts[0:4,:,:]+counts[4:8,:,:] # combine f and r ATCG counts

    sorted_c = np.sort(c,axis=0) # sort by num. ATCGs 
    argsort_c = np.argsort(c,axis=0)
    
    # Get allele counts for major allele (4th row)
    # Weird "3:4:" indexing required to maintain 3D structure
    maxcount = sorted_c[3:4:,:,:] 
    # Get allele counts for first minor allele (3rd row)
    # tri/quadro-allelic ignored!!
    minorcount = sorted_c[2:3:,:,:] 
    
    with np.errstate(divide='ignore', invalid='ignore'):
        
        maf = maxcount / sorted_c.sum(axis=0,keepdims=True)
        minorAF = minorcount / sorted_c.sum(axis=0,keepdims=True)
    
    # turn 2D; axis=1 to keep 2d structure when only one position!
    maf = np.squeeze(maf,axis=0) 
    maf[np.isnan(maf)]=0 # set to 0 to indicate no data
    
    minorAF = np.squeeze(minorAF,axis=0) 
    minorAF[np.isnan(minorAF)]=0 # set to 0 to indicate no data/no minor AF
    
    # Idx given by argsort_c represents allele position ATCG
    # A=0,T=1,C=2,G=3
    # axis=1 to keep 2d structure when only one position!
    maNT = np.squeeze(argsort_c[3:4:,:,:],axis=0) 
    minorNT = np.squeeze(argsort_c[2:3:,:,:],axis=0)

    # Note: If counts for all bases are zero, then sort won't change the order
    # (since there is nothing to sort), thus maNT/minorNT will be put to -1 (NA)
    # using maf (REMEMBER: minorAF==0 is a value!)
    maNT[maf==0]=-1
    minorNT[maf==0]=-1
    
    # MATLAB conversion to NATCG=01234
    # !Important! This is required for current ver of find_clade_specific_snps
    # as of 3/26/22; Want to change later
    maNT=maNT+1
    minorNT=minorNT+1
    
    return maNT, maf, minorNT, minorAF

#To do: simplify into numpy array
#Resolve whether I need sample_names or not
def distmat(calls, sample_names):
    ''' Calculate the pairwise SNP distance of all samples in a maNT matrix.

    Args:
        calls (arr): Matrix of major allele NT for each sample.
        sample_names (ls): List of sample names.

    Returns:
        distmat (arr): Matrix of pairwise distances.

    '''
    num_samples=len(sample_names)
    
    distmat = np.zeros((num_samples,num_samples))
    
    for i in range(num_samples):
        # print(f"Sample progress: {i+1}/{num_samples} samples done ")
        distmat[i,:] = np.count_nonzero( (calls != np.tile(calls[:,i],(num_samples,1)).T) &
                               (calls > 0) &
                               (np.tile(calls[:,i],(num_samples,1)).T > 0) , axis=0)

    distmat_df = pd.DataFrame(distmat, 
                          index=sample_names, 
                          columns=sample_names)
    
    return distmat_df


def p2chrpos(p, ChrStarts):
    '''# return 2col array with chr and pos on chr
    #p --> continous, ignores chr
    #pos --> like p, 0-based'''

    # get chr and pos-on-chr
    chr = np.ones(len(p),dtype=int)
    if len(ChrStarts) > 1:
        for i in ChrStarts[1:]:
            chr = chr + (p > i) # when (p > i) evaluates 'true' lead to plus 1 in summation. > bcs ChrStarts start with 0...genomestats()
        positions = p - ChrStarts[chr-1] # [chr-1] -1 due to 0based index
        pos = np.column_stack((chr,positions))
    else:
        pos = np.column_stack((chr,p))
    return pos

# To do: format
def idx2nts(calls, missingdata="?"):
    # translate index array to array containing nucleotides
    # add 5th element --> no data! == index -1
    nucl = np.array([missingdata,'A','T','C','G'],dtype=object) 
    palette = [-1,0,1,2,3] # values present in index-array
    index = np.digitize(calls.ravel(), palette, right=True)
    
    return nucl[index].reshape(calls.shape)

# To do: format
def write_calls_to_fasta(calls,sample_names,output_file):
    
    fa_file = open(output_file, "w")
    
    for i,name in enumerate(sample_names):
        nucl_string = "".join(list(calls[:,i]))
        fa_file.write(">" + name + "\n" + nucl_string + "\n")
    
    fa_file.close()    

def read_fasta(reference_genome_file): 
    '''Reads in fasta file. If directory is given, reads in dir/genome.fasta
    Args:
        reference_genome_file (str): Path to reference genome.

    Returns: SeqIO object for reference genome.
    '''

    if os.path.exists(reference_genome_file):
        refgenome = SeqIO.parse(reference_genome_file,'fasta')

        return refgenome

    fasta_file = glob.glob(reference_genome_file + '/genome.fasta')
    if len(fasta_file) != 1:
        fasta_file_gz = glob.glob(reference_genome_file + '/genome.fasta.gz')
        if len(fasta_file_gz) != 1:
            raise ValueError('Either no genome.fasta(.gz) or more than 1 genome.fasta(.gz) file found in ' + reference_genome_file)
        else: # genome.fasta.gz
            refgenome = SeqIO.parse(gzip.open(fasta_file_gz[0], "rt"),'fasta')
    else: # genome.fasta
        refgenome = SeqIO.parse(fasta_file[0],'fasta')
    
    return refgenome

def genomestats(path_to_refgenome):
    '''Parse genome to extract relevant stats

    Args:
        REFGENOMEFOLDER (str): Path to reference genome.

    Returns:
        ChrStarts (arr): DESCRIPTION.
        Genomelength (arr): DESCRIPTION.
        ScafNames (arr): DESCRIPTION.

    '''

    refgenome = read_fasta(path_to_refgenome)
    
    Genomelength = 0
    ChrStarts = []
    ScafNames = []
    for record in refgenome:
        ChrStarts.append(Genomelength) # chr1 starts at 0 in analysis.m
        Genomelength = Genomelength + len(record)
        ScafNames.append(record.id)
    # close file
    #refgenome.close() # biopy update SeqIO has no close attribute anymore.
    # turn to np.arrys!
    ChrStarts = np.asarray(ChrStarts,dtype=int)
    Genomelength = np.asarray(Genomelength,dtype=int)
    ScafNames = np.asarray(ScafNames,dtype=object)
    
    return ChrStarts,Genomelength,ScafNames

def p2chrpos(p, ChrStarts):
    '''Convert 1col list of pos to 2col array with chromosome and pos on chromosome

    Args:
        p (TYPE): DESCRIPTION.
        ChrStarts (TYPE): DESCRIPTION.

    Returns:
        chrpos (TYPE): DESCRIPTION.

    '''
        
    # get chr and pos-on-chr
    chromo = np.ones(len(p),dtype=int)
    if len(ChrStarts) > 1:
        for i in ChrStarts[1:]:
            chromo = chromo + (p > i) # when (p > i) evaluates 'true' lead to plus 1 in summation. > bcs ChrStarts start with 0...genomestats()
        positions = p - ChrStarts[chromo-1] # [chr-1] -1 due to 0based index
        chrpos = np.column_stack((chromo,positions))
    else:
        chrpos = np.column_stack((chromo,p))
    return chrpos

def parse_file_list(path_to_file):
    with open(path_to_file, 'r') as f:
        file_list = f.read().splitlines()

    for file_ in file_list:
        if not os.path.isabs(file_):
            abs_path = os.path.abspath(file_)
            if not os.path.exists(abs_path):
                raise FileNotFoundError(f"File not found: {abs_path}")
            file_list[file_list.index(file_)] = os.path.abspath(file_)

    return file_list
